# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import asyncio
import collections.abc
import contextlib
import json
import os
import socket
import typing
import unittest
from unittest.mock import Mock

from asyncvarlink import (
    FileDescriptor,
    override,
    TypedVarlinkErrorReply,
    VarlinkInterface,
    VarlinkInterfaceRegistry,
    VarlinkInterfaceServerProtocol,
    VarlinkTransport,
    varlinkmethod,
)

from helpers import async_read_fd, async_send_fds, defer


class DemoError(TypedVarlinkErrorReply, interface="com.example.demo"):
    class Parameters:
        pass


class DemoInterface(VarlinkInterface, name="com.example.demo"):
    def __init__(self, fut: asyncio.Future[int]):
        super().__init__()
        self.fut = fut

    @varlinkmethod(return_parameter="result")
    def Answer(self) -> int:
        return 42

    @varlinkmethod
    def Error(self) -> None:
        raise DemoError()

    @varlinkmethod
    async def AsyncError(self) -> None:
        await asyncio.sleep(0)
        raise DemoError()

    @varlinkmethod(return_parameter="result")
    async def FutureAnswer(self) -> int:
        return await self.fut

    @varlinkmethod(return_parameter="result")
    def SyncMore(self) -> collections.abc.Iterator[int]:
        yield 1
        yield 2

    @varlinkmethod(return_parameter="result")
    def SyncMoreError(self) -> collections.abc.Iterator[int]:
        yield 1
        raise DemoError()

    @varlinkmethod(return_parameter="result")
    async def AsyncMore(self) -> collections.abc.AsyncIterator[int]:
        yield 1
        yield 2

    @varlinkmethod(return_parameter="fd")
    async def CreateFd(
        self, kind: typing.Literal["pipe", "socket"]
    ) -> FileDescriptor:
        if kind == "pipe":
            rend, wend = os.pipe()
            os.write(wend, b"needle")
            os.close(wend)
            return FileDescriptor(rend, True)
        assert kind == "socket"
        sock1, sock2 = socket.socketpair()
        sock1.send(b"needle")
        sock1.close()
        return FileDescriptor(sock2, True)

    @varlinkmethod(return_parameter="content")
    async def ReadFromSocket(self, fd: FileDescriptor) -> str:
        fileno = fd.take()
        assert fileno is not None
        if not isinstance(fileno, int):
            fileno = fileno.fileno()
        with socket.socket(fileno=fileno) as sock:
            sock.setblocking(False)
            data = await asyncio.get_running_loop().sock_recv(sock, 1024)
            return data.decode("utf8")


class ServerTests(unittest.IsolatedAsyncioTestCase):
    @override
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.registry = VarlinkInterfaceRegistry()
        self.fut = asyncio.get_running_loop().create_future()
        self.registry.register_interface(DemoInterface(self.fut))

    @contextlib.asynccontextmanager
    async def connected_server(
        self,
    ) -> collections.abc.AsyncIterator[tuple[socket.socket, socket.socket]]:
        loop = asyncio.get_running_loop()
        sock1, sock2 = socket.socketpair(
            type=socket.SOCK_STREAM | socket.SOCK_NONBLOCK
        )
        self.protocol = VarlinkInterfaceServerProtocol(self.registry)
        transport: VarlinkTransport | None = None
        try:
            transport = VarlinkTransport(loop, sock2, sock2, self.protocol)
            yield (sock1, sock2)
        finally:
            if transport:
                transport.close()
                await asyncio.sleep(0)
                self.assertLess(sock2.fileno(), 0)
            else:
                sock2.close()
            sock1.close()

    @contextlib.asynccontextmanager
    async def piped_server(
        self,
    ) -> collections.abc.AsyncIterator[tuple[FileDescriptor, FileDescriptor]]:
        loop = asyncio.get_running_loop()
        self.protocol = VarlinkInterfaceServerProtocol(self.registry)
        pipe1r, pipe1w = FileDescriptor.make_pipe()
        pipe2r, pipe2w = FileDescriptor.make_pipe()
        # FileDescriptor makes close idempotent
        with (
            contextlib.closing(pipe1r),
            contextlib.closing(pipe1w),
            contextlib.closing(pipe2r),
            contextlib.closing(pipe2w),
        ):
            transport: VarlinkTransport | None = None
            try:
                transport = VarlinkTransport(
                    loop, pipe2r, pipe1w, self.protocol
                )
                yield (pipe1r, pipe2w)
            finally:
                if transport:
                    transport.close()
                    await asyncio.sleep(0)
                    self.assertFalse(pipe2r)
                    self.assertFalse(pipe1w)

    def sock_recv_fds(
        self, sock: socket.socket
    ) -> asyncio.Future[tuple[bytes, list[int]]]:
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        loop.add_reader(sock, self._sock_recv_fds, sock, fut)
        return fut

    def _sock_recv_fds(
        self, sock: socket.socket, fut: asyncio.Future[tuple[bytes, list[int]]]
    ) -> None:
        asyncio.get_running_loop().remove_reader(sock)
        try:
            data, fds, _flags, _addr = socket.recv_fds(sock, 1024, 32)
        except Exception as exc:
            fut.set_exception(exc)
        else:
            fut.set_result((data, fds))

    async def invoke(self, request: bytes, expected_response: bytes) -> None:
        loop = asyncio.get_running_loop()
        async with self.connected_server() as (sock1, _):
            await loop.sock_sendall(sock1, request + b"\0")
            data = await loop.sock_recv(sock1, 1024)
            if json.loads(data.split(b"\0", 1)[0]).get("continues"):
                data += await loop.sock_recv(sock1, 1024)
            self.assertEqual(data, expected_response + b"\0")

    async def test_sync_single(self) -> None:
        await self.invoke(
            b'{"method":"com.example.demo.Answer"}',
            b'{"parameters":{"result":42}}',
        )

    async def test_more(self) -> None:
        await self.invoke(
            b'{"method":"com.example.demo.SyncMore","more":true}',
            b'{"continues":true,"parameters":{"result":1}}\0'
            b'{"parameters":{"result":2}}',
        )
        await self.invoke(
            b'{"method":"com.example.demo.AsyncMore","more":true}',
            b'{"continues":true,"parameters":{"result":1}}\0'
            b'{"parameters":{"result":2}}',
        )

    async def test_error(self) -> None:
        await self.invoke(
            b'{"method":"com.example.demo.Error"}',
            b'{"error":"com.example.demo.DemoError"}',
        )
        await self.invoke(
            b'{"method":"com.example.demo.SyncMore"}',
            b'{"error":"org.varlink.service.ExpectedMore"}',
        )

    async def test_async_error(self) -> None:
        await self.invoke(
            b'{"method":"com.example.demo.AsyncError"}',
            b'{"error":"com.example.demo.DemoError"}',
        )

    async def test_more_error(self) -> None:
        await self.invoke(
            b'{"method":"com.example.demo.SyncMoreError","more":true}',
            b'{"continues":true,"parameters":{"result":1}}\0'
            b'{"error":"com.example.demo.DemoError"}',
        )

    async def test_invalid_interface(self) -> None:
        await self.invoke(
            b'{"method":"com.example.doesnotexist.Anything"}',
            b'{"error":"org.varlink.service.InterfaceNotFound","parameters":{"interface":"com.example.doesnotexist"}}',
        )

    async def test_invalid_method(self) -> None:
        await self.invoke(
            b'{"method":"com.example.demo.DoesNotExist"}',
            b'{"error":"org.varlink.service.MethodNotFound","parameters":{"method":"DoesNotExist"}}',
        )

    async def test_invalid_parameters(self) -> None:
        await self.invoke(
            b'{"method":"com.example.demo.Answer","parameters":{"unexpected":1}}',
            b'{"error":"org.varlink.service.InvalidParameter","parameters":{"parameter":"unexpected"}}',
        )

    async def test_async(self) -> None:
        self.fut.set_result(42)
        await self.invoke(
            b'{"method":"com.example.demo.FutureAnswer"}',
            b'{"parameters":{"result":42}}',
        )

    async def test_return_fd(self) -> None:
        loop = asyncio.get_running_loop()
        async with self.connected_server() as (sock1, sock2):
            for kind in ("pipe", "socket"):
                with self.subTest(kind=kind):
                    await loop.sock_sendall(
                        sock1,
                        b'{"method":"com.example.demo.CreateFd","parameters":{"kind":"%s"}}\0'
                        % kind.encode("ascii"),
                    )
                    data, fds = await self.sock_recv_fds(sock1)
                    self.assertEqual(data, b'{"parameters":{"fd":0}}\0')
                    self.assertEqual(len(fds), 1)
                    data = await async_read_fd(fds[0], 1024)
                    os.close(fds[0])
                    self.assertEqual(data, b"needle")

    async def test_pass_fd(self) -> None:
        loop = asyncio.get_running_loop()
        async with self.connected_server() as (clsock, srvsock):
            writesock, readsock = socket.socketpair(
                type=socket.SOCK_STREAM | socket.SOCK_NONBLOCK
            )
            self.addCleanup(writesock.close)
            self.addCleanup(readsock.close)
            await loop.sock_sendall(writesock, b"spam")
            # This is blocking in theory, but it practically does not block.
            await async_send_fds(
                clsock,
                b'{"method":"com.example.demo.ReadFromSocket","parameters":{"fd": 0}}\0',
                [readsock.fileno()],
            )
            data, fds = await self.sock_recv_fds(clsock)
            self.assertEqual(data, b'{"parameters":{"content":"spam"}}\0')
            self.assertEqual(fds, [])

    async def test_protocol_violation(self) -> None:
        await self.invoke(
            b"{}",
            b'{"error":"invalid.asyncvarlink.ProtocolViolation"}',
        )

    async def test_broken_socket(self) -> None:
        loop = asyncio.get_running_loop()
        async with self.connected_server() as (sock1, sock2):
            self.protocol.connection_lost = Mock(return_value=None)
            await loop.sock_sendall(
                sock1, b'{"method":"com.example.demo.FutureAnswer"}\0'
            )
            sock1.close()
            await defer(count=10)
            # Since the future is still pending, the write end is not being
            # closed yet.
            self.assertFalse(self.protocol.connection_lost.called)
            self.fut.set_result(42)
            await defer(until_called=self.protocol.connection_lost)
            self.protocol.connection_lost.assert_called_once_with(None)
            self.assertLess(sock2.fileno(), 0)

    async def test_pipe_closed(self) -> None:
        loop = asyncio.get_running_loop()
        async with self.piped_server() as (rpipe, wpipe):
            self.protocol.connection_lost = Mock(return_value=None)
            os.write(
                wpipe.fileno(), b'{"method":"com.example.demo.FutureAnswer"}\0'
            )
            wpipe.close()
            await defer(count=10)
            # Since the future is still pending, the write end is not being
            # closed yet.
            self.assertFalse(self.protocol.connection_lost.called)
            self.fut.set_result(42)
            data = await async_read_fd(rpipe.fileno(), 1024)
            self.assertEqual(data, b'{"parameters":{"result":42}}\0')
            await defer(until_called=self.protocol.connection_lost)
            self.protocol.connection_lost.assert_called_once_with(None)

    async def test_broken_pipe(self) -> None:
        loop = asyncio.get_running_loop()
        async with self.piped_server() as (rpipe, wpipe):
            self.protocol.connection_lost = Mock(return_value=None)
            rpipe.close()
            os.write(
                wpipe.fileno(), b'{"method":"com.example.demo.FutureAnswer"}\0'
            )
            wpipe.close()
            self.fut.set_result(42)
            await defer(until_called=self.protocol.connection_lost)
            self.protocol.connection_lost.assert_called_once_with(None)



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import contextlib
import tempfile
import unittest

from asyncvarlink import (
    VarlinkClientProtocol,
    VarlinkInterface,
    VarlinkInterfaceRegistry,
    connect_unix_varlink,
    create_unix_server,
    varlinkmethod,
)


class DummyInterface(VarlinkInterface, name="com.example.Dummy"):
    def __init__(self) -> None:
        self.argument = "unset"

    @varlinkmethod(return_parameter="result")
    def Method(self, argument: str) -> str:
        self.argument = argument
        return "returnvalue"


class End2EndTests(unittest.IsolatedAsyncioTestCase):
    async def test_end2end(self) -> None:
        registry = VarlinkInterfaceRegistry()
        interface = DummyInterface()
        registry.register_interface(interface)
        async with contextlib.AsyncExitStack() as stack:
            tdir = stack.enter_context(tempfile.TemporaryDirectory())
            sockpath = tdir + "/sock"
            server = await stack.enter_async_context(
                await create_unix_server(registry.protocol_factory, sockpath)
            )
            stack.callback(server.close)
            transport, protocol = await connect_unix_varlink(
                VarlinkClientProtocol, sockpath
            )
            stack.callback(transport.close)
            assert isinstance(protocol, VarlinkClientProtocol)
            proxy = protocol.make_proxy(DummyInterface)
            self.assertEqual(
                await proxy.Method(argument="argument"),
                {"result": "returnvalue"},
            )
            self.assertEqual(interface.argument, "argument")



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""asyncio varlink server protocol implementation"""

import asyncio
import collections.abc
import functools
import logging
import os
import typing

from .conversion import ConversionError, FileDescriptorVarlinkType
from .error import VarlinkErrorReply, GenericVarlinkErrorReply
from .interface import (
    AnnotatedResult,
    VarlinkInterface,
    VarlinkMethodSignature,
    varlinksignature,
)
from .message import VarlinkMethodCall, VarlinkMethodReply
from .protocol import VarlinkBaseProtocol, VarlinkProtocol
from .serviceerrors import (
    ExpectedMore,
    InterfaceNotFound,
    InvalidParameter,
    MethodNotFound,
)
from .types import FileDescriptorArray, JSONObject, override


_logger = logging.getLogger("asyncvarlink.serverprotocol")


class VarlinkInterfaceRegistry:
    """Collection of VarlinkInterface instances."""

    def __init__(self) -> None:
        self.interfaces: dict[str, VarlinkInterface] = {}

    def register_interface(self, interface: VarlinkInterface) -> None:
        """Register an interface instance. Its name must be unique to the
        registry.
        """
        if interface.name in self.interfaces:
            raise ValueError(
                f"an interface named {interface.name} is already registered"
            )
        self.interfaces[interface.name] = interface

    def lookup_method(
        self, call: VarlinkMethodCall
    ) -> tuple[
        collections.abc.Callable[..., typing.Any], VarlinkMethodSignature
    ]:
        """Look up a method. Return the Python callable responsible for the
        method referenced by the call and its VarlinkMethodSignature used
        for introspection and type conversion. This raises a number of
        subclasses of VarlinkErrorReply.
        """
        try:
            interface = self.interfaces[call.method_interface]
        except KeyError:
            raise InterfaceNotFound(interface=call.method_interface) from None
        try:
            method = getattr(interface, call.method_name)
        except AttributeError:
            raise MethodNotFound(method=call.method_name) from None
        if (signature := varlinksignature(method)) is None:
            # Reject any method that has not been marked with varlinkmethod.
            raise MethodNotFound(method=call.method_name)
        if signature.more and not call.more:
            raise ExpectedMore()
        return (method, signature)

    def __iter__(self) -> collections.abc.Iterator[VarlinkInterface]:
        """Iterate over the registered VarlinkInterface instances."""
        return iter(self.interfaces.values())

    def __getitem__(self, interface: str) -> VarlinkInterface:
        """Look up a VarlinkInterface by its name. Raises KeyError."""
        return self.interfaces[interface]

    def protocol_factory(self) -> VarlinkBaseProtocol:
        """Factory method for generating protocol instances.
        Example:

            create_unix_server(registry.protocol_factory, ...)
        """
        return VarlinkInterfaceServerProtocol(self)


class VarlinkServerProtocol(VarlinkProtocol):
    """Protocol class for a varlink service. It receives calls as
    VarlinkMethodCall objects and issues replies as VarlinkMethodReply or
    VarlinkErrorReply objects. A derived class should implement call_received.
    """

    def __init__(self) -> None:
        super().__init__()
        self._inflight: set[asyncio.Future[None]] = set()
        self._closing = False

    def _maybe_close_if_idle(self) -> None:
        if self._closing:
            while self._inflight:
                fut = next(iter(self._inflight))
                if not fut.done():
                    return
                self._inflight.remove(fut)
            self.close()
            self._closing = False

    def _on_transfer_done(self, fut: asyncio.Future[None]) -> None:
        assert fut.done()
        self._inflight.discard(fut)
        self._maybe_close_if_idle()

    def _add_transfer(self, fut: asyncio.Future[None]) -> None:
        self._inflight.add(fut)
        fut.add_done_callback(self._on_transfer_done)

    def _on_message_sent(self, fut: asyncio.Future[None]) -> None:
        assert fut.done()
        exc = fut.exception()
        if exc is None or isinstance(exc, BrokenPipeError):
            # The lower layer will close the connection on BrokenPipeError.
            return
        _logger.exception("Sending reply failed", exc_info=exc)

    def send_reply(
        self,
        reply: VarlinkMethodReply,
        fds: list[int] | None = None,
        autoclose: bool = True,
    ) -> asyncio.Future[None]:
        """Enqueue the given reply and file descriptors for sending. For the
        semantics regarding fds, please refer to the documentation of
        send_message.
        """
        fut = self.send_message(reply.tojson(), fds, autoclose)
        fut.add_done_callback(self._on_message_sent)
        self._add_transfer(fut)
        return fut

    def send_error(self, error: VarlinkErrorReply) -> asyncio.Future[None]:
        """Enqueue the given error for sending. The future shall be done when
        the error has been sent.
        """
        try:
            reply = error.toreply()
        except ConversionError as err:
            reply = InvalidParameter(parameter=err.location[0]).toreply()
        return self.send_reply(reply)

    def _on_receiver_completes(
        self,
        backpressure_fut: asyncio.Future[None],
        oneway: bool,
        call_fut: asyncio.Future[None],
    ) -> None:
        assert call_fut.done()
        exc = call_fut.exception()
        if exc is not None:
            if isinstance(exc, VarlinkErrorReply):
                if not oneway:
                    self.send_error(exc)
            elif isinstance(exc, (SystemExit, KeyboardInterrupt)):
                raise exc
            else:
                try:
                    self.handle_call_exception(exc, oneway)
                except VarlinkErrorReply as err:
                    if not oneway:
                        self.send_error(err)
        backpressure_fut.set_result(None)

    def handle_call_exception(self, exc: BaseException, oneway: bool) -> None:
        """React to call_received emitting an exception not inheriting from
        VarlinkErrorReply. The caught exception is provided. The default
        implementation logs the exception and sends an error reply unless
        oneway is True. The function may raise a VarlinkErrorReply and have it
        sent when needed.
        """
        _logger.exception(
            "unhandled exception from call_received future", exc_info=exc
        )
        if not oneway:
            self.send_error(
                GenericVarlinkErrorReply(
                    "invalid.asyncvarlink.InternalServerError"
                )
            )

    @override
    def request_received(
        self, obj: JSONObject, fds: FileDescriptorArray | None
    ) -> asyncio.Future[None] | None:
        try:
            try:
                call = VarlinkMethodCall.fromjson(obj)
            except (TypeError, ValueError):
                raise GenericVarlinkErrorReply(
                    "invalid.asyncvarlink.ProtocolViolation"
                ) from None
            fut = self.call_received(call, fds)
            if fut is None:
                return None
            bpfut = asyncio.get_running_loop().create_future()
            fut.add_done_callback(
                functools.partial(
                    self._on_receiver_completes, bpfut, call.oneway
                ),
            )
            self._add_transfer(fut)
            return bpfut
        except VarlinkErrorReply as err:
            if not obj.get("oneway", False):
                self.send_error(err)
            return None
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as exc:
            oneway = bool(obj.get("oneway", False))
            try:
                self.handle_call_exception(exc, oneway)
            except VarlinkErrorReply as err:
                if not oneway:
                    self.send_error(err)
            return None

    @override
    def eof_received(self) -> None:
        super().eof_received()
        self._closing = True
        self._maybe_close_if_idle()

    @override
    def connection_lost(self, exc: Exception | None) -> None:
        super().connection_lost(exc)
        self.close()

    def call_received(
        self, call: VarlinkMethodCall, fds: FileDescriptorArray | None
    ) -> asyncio.Future[None] | None:
        """Handle a received varlink parsed as a call object and associated
        file descriptors. The descriptors are valid until the function returns.
        Their life time can be extended by adding a referee before returning.
        If the function raises an exception or the returned future (if any)
        completes with an exception a reply will be generated. Otherwise, the
        function is expected to call send_reply as needed before returning None
        or completing the returned future.
        """
        raise NotImplementedError


class VarlinkInterfaceServerProtocol(VarlinkServerProtocol):
    """Serve the interfaces registered with a registry via varlink."""

    def __init__(self, registry: VarlinkInterfaceRegistry) -> None:
        """Defer method lookup to the given registry."""
        super().__init__()
        self._registry = registry

    @override
    def call_received(
        self, call: VarlinkMethodCall, fds: FileDescriptorArray | None
    ) -> asyncio.Future[None] | None:
        method, signature = self._registry.lookup_method(call)
        try:
            pyparams = signature.parameter_type.fromjson(
                call.parameters, {FileDescriptorVarlinkType: fds}
            )
        except ConversionError as err:
            raise InvalidParameter(parameter=err.location[0]) from err
        if not signature.asynchronous:
            if signature.more:
                fut = asyncio.ensure_future(
                    self._call_sync_method_more(method, signature, pyparams)
                )
            else:
                self._call_sync_method_single(
                    method, signature, pyparams, call.oneway
                )
                return None
        elif signature.more:
            fut = asyncio.ensure_future(
                self._call_async_method_more(method, signature, pyparams)
            )
        else:
            fut = asyncio.ensure_future(
                self._call_async_method_single(
                    method, signature, pyparams, call.oneway
                ),
            )
        if fds is not None:
            fds.reference_until_done(fut)
        return fut

    def _call_sync_method_single(
        self,
        method: collections.abc.Callable[..., typing.Any],
        signature: VarlinkMethodSignature,
        pyparams: dict[str, typing.Any],
        oneway: bool,
    ) -> asyncio.Future[None] | None:
        result = method(**pyparams)
        assert isinstance(result, AnnotatedResult)
        assert not result.continues
        if oneway:
            return None
        with FileDescriptorArray.new_managed() as fds:
            jsonparams = signature.return_type.tojson(
                result.parameters, {FileDescriptorVarlinkType: fds}
            )
            fut = self.send_reply(
                VarlinkMethodReply(jsonparams, extensions=result.extensions),
                [fd.fileno() for fd in fds],
                autoclose=False,
            )
            fds.reference_until_done(fut)
        return fut

    async def _call_sync_method_more(
        self,
        method: collections.abc.Callable[..., typing.Any],
        signature: VarlinkMethodSignature,
        pyparams: dict[str, typing.Any],
    ) -> None:
        try:
            continues = True
            generator = method(**pyparams)
            for result in generator:
                assert continues
                assert isinstance(result, AnnotatedResult)
                with FileDescriptorArray.new_managed() as fds:
                    jsonparams = signature.return_type.tojson(
                        result.parameters, {FileDescriptorVarlinkType: fds}
                    )
                    fut = self.send_reply(
                        VarlinkMethodReply(
                            jsonparams,
                            continues=result.continues,
                            extensions=result.extensions,
                        ),
                        [fd.fileno() for fd in fds],
                        autoclose=False,
                    )
                    fds.reference_until_done(fut)
                try:
                    await fut
                except OSError:
                    generator.close()
                    return
                continues = result.continues
            assert not continues
        except VarlinkErrorReply as err:
            self.send_error(err)

    async def _call_async_method_single(
        self,
        method: collections.abc.Callable[..., typing.Any],
        signature: VarlinkMethodSignature,
        pyparams: dict[str, typing.Any],
        oneway: bool,
    ) -> None:
        try:
            result = await method(**pyparams)
            assert isinstance(result, AnnotatedResult)
            assert not result.continues
            if oneway:
                return
            with FileDescriptorArray.new_managed() as fds:
                jsonparams = signature.return_type.tojson(
                    result.parameters, {FileDescriptorVarlinkType: fds}
                )
                fut = self.send_reply(
                    VarlinkMethodReply(
                        jsonparams, extensions=result.extensions
                    ),
                    [fd.fileno() for fd in fds],
                    autoclose=False,
                )
                fds.reference_until_done(fut)
        except VarlinkErrorReply as err:
            if not oneway:
                self.send_error(err)

    async def _call_async_method_more(
        self,
        method: collections.abc.Callable[..., typing.Any],
        signature: VarlinkMethodSignature,
        pyparams: dict[str, typing.Any],
    ) -> None:
        try:
            continues = True
            generator = method(**pyparams)
            async for result in generator:
                assert continues
                assert isinstance(result, AnnotatedResult)
                with FileDescriptorArray.new_managed() as fds:
                    jsonparams = signature.return_type.tojson(
                        result.parameters, {FileDescriptorVarlinkType: fds}
                    )
                    fut = self.send_reply(
                        VarlinkMethodReply(
                            jsonparams,
                            continues=result.continues,
                            extensions=result.extensions,
                        ),
                        [fd.fileno() for fd in fds],
                        autoclose=False,
                    )
                    fds.reference_until_done(fut)
                try:
                    await fut
                except OSError:
                    generator.aclose()
                    return
                continues = result.continues
            assert not continues
        except VarlinkErrorReply as err:
            self.send_error(err)



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import asyncio
import collections.abc
import json
import os
import socket
import unittest

from asyncvarlink import (
    ConversionError,
    FileDescriptor,
    override,
    TypedVarlinkErrorReply,
    VarlinkClientProtocol,
    VarlinkInterface,
    VarlinkTransport,
    varlinkmethod,
)
from asyncvarlink.serviceinterface import (
    PermissionDenied,
    VarlinkServiceInterface,
)

from helpers import async_read_fd, async_send_fds


class DemoFailure(TypedVarlinkErrorReply):
    name = "com.example.demo.DemoFailure"

    class Parameters:
        pass


class DemoInterface(VarlinkInterface, name="com.example.demo"):
    errors = (DemoFailure,) + VarlinkServiceInterface.errors

    @varlinkmethod(return_parameter="result")
    def Method(self, argument: str) -> str: ...

    @varlinkmethod(return_parameter="result")
    def MoreMethod(self) -> collections.abc.Iterator[str]: ...

    @varlinkmethod(return_parameter="fd")
    def CreateFd(self) -> FileDescriptor: ...


class ClientTests(unittest.IsolatedAsyncioTestCase):
    @override
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.loop = asyncio.get_running_loop()
        self.sock1, self.sock2 = socket.socketpair(
            type=socket.SOCK_STREAM | socket.SOCK_NONBLOCK
        )
        self.proto = VarlinkClientProtocol()
        self.transport = VarlinkTransport(
            self.loop, self.sock2, self.sock2, self.proto
        )
        self.proxy = self.proto.make_proxy(DemoInterface)

    @override
    async def asyncTearDown(self) -> None:
        self.transport.close()
        await asyncio.sleep(0)
        self.assertLess(self.sock2.fileno(), 0)
        self.sock1.close()
        await super().asyncTearDown()

    async def expect_data(self, expected: bytes) -> None:
        data = await self.loop.sock_recv(self.sock1, len(expected) + 1)
        self.assertEqual(data, expected)

    async def test_simple(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await self.expect_data(
            b'{"method":"com.example.demo.Method","parameters":{"argument":"spam"}}\0'
        )
        self.assertFalse(fut.done())
        await async_send_fds(self.sock1, b'{"parameters":{"result":"egg"}}\0')
        self.assertEqual(await fut, {"result": "egg"})

    async def test_more(self) -> None:
        gen = self.proxy.MoreMethod()
        fut = asyncio.ensure_future(anext(gen))
        await self.expect_data(
            b'{"method":"com.example.demo.MoreMethod","more":true}\0'
        )
        self.assertFalse(fut.done())
        await async_send_fds(
            self.sock1, b'{"continues":true,"parameters":{"result":"spam"}}\0'
        )
        self.assertEqual(await fut, {"result": "spam"})
        fut = asyncio.ensure_future(anext(gen))
        await asyncio.sleep(0)
        self.assertFalse(fut.done())
        await async_send_fds(self.sock1, b'{"parameters":{"result":"egg"}}\0')
        self.assertEqual(await fut, {"result": "egg"})
        with self.assertRaises(StopAsyncIteration):
            await anext(gen)

    async def test_fragmentation(self) -> None:
        fut1 = asyncio.ensure_future(self.proxy.Method(argument="first"))
        await self.expect_data(
            b'{"method":"com.example.demo.Method","parameters":{"argument":"first"}}\0'
        )
        await async_send_fds(self.sock1, b'{"parameters":')
        fut2 = asyncio.ensure_future(self.proxy.Method(argument="second"))
        await self.expect_data(
            b'{"method":"com.example.demo.Method","parameters":{"argument":"second"}}\0'
        )
        self.assertFalse(fut1.done())
        self.assertFalse(fut2.done())
        await async_send_fds(
            self.sock1, b'{"result":"one"}}\0{"parameters":{"result":"two"}}\0'
        )
        self.assertEqual(await fut1, {"result": "one"})
        self.assertEqual(await fut2, {"result": "two"})

    async def test_invalid_argument(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(invalid_argument=True))
        await asyncio.sleep(0)
        self.assertTrue(fut.done())
        self.assertRaises(ConversionError, fut.result)

    async def test_demo_error(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await self.expect_data(
            b'{"method":"com.example.demo.Method","parameters":{"argument":"spam"}}\0'
        )
        self.assertFalse(fut.done())
        await async_send_fds(
            self.sock1, b'{"error":"com.example.demo.DemoFailure"}\0'
        )
        with self.assertRaises(DemoFailure):
            await fut

    async def test_permission_denied(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await self.expect_data(
            b'{"method":"com.example.demo.Method","parameters":{"argument":"spam"}}\0'
        )
        self.assertFalse(fut.done())
        await async_send_fds(
            self.sock1, b'{"error":"org.varlink.service.PermissionDenied"}\0'
        )
        with self.assertRaises(PermissionDenied):
            await fut

    async def test_broken_pipe_send(self) -> None:
        self.sock1.close()
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        with self.assertRaises(BrokenPipeError):
            await fut
        self.assertLess(self.sock2.fileno(), 0)

    async def test_broken_pipe_receive(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await self.expect_data(
            b'{"method":"com.example.demo.Method","parameters":{"argument":"spam"}}\0'
        )
        self.assertFalse(fut.done())
        self.sock1.close()
        with self.assertRaises(ConnectionResetError):
            await fut
        self.assertLess(self.sock2.fileno(), 0)

    async def test_non_json_reply(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await async_send_fds(self.sock1, b"this is not json\0")
        with self.assertRaises(json.JSONDecodeError):
            await fut

    async def test_non_object_reply(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await async_send_fds(self.sock1, b'"not an object"\0')
        with self.assertRaises(TypeError):
            await fut

    async def test_missing_fields(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await async_send_fds(self.sock1, b'{"missing fields":0}\0')
        with self.assertRaises(ConversionError):
            await fut

    async def test_return_fd(self) -> None:
        fut = asyncio.ensure_future(self.proxy.CreateFd())
        await self.expect_data(b'{"method":"com.example.demo.CreateFd"}\0')
        self.assertFalse(fut.done())
        rend, wend = os.pipe()
        self.addCleanup(os.close, rend)
        read_fut = async_read_fd(rend, 7)
        await async_send_fds(self.sock1, b'{"parameters":{"fd":0}}\0', [wend])
        with await fut as ret:
            wfd = ret["fd"].take()
        self.addCleanup(os.close, wfd)
        os.close(wend)  # Defer closure for assertNotEqual
        self.assertNotEqual(wfd, wend)
        self.assertFalse(read_fut.done())
        os.write(wfd, b"needle")
        self.assertEqual(await read_fut, b"needle")


class ClientPipeTests(unittest.IsolatedAsyncioTestCase):
    @override
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.loop = asyncio.get_running_loop()
        self.pipers, self.pipewc = os.pipe()
        self.piperc, self.pipews = os.pipe()
        os.set_blocking(self.pipers, False)
        os.set_blocking(self.pipews, False)
        self.proto = VarlinkClientProtocol()
        self.transport = VarlinkTransport(
            self.loop, self.piperc, self.pipewc, self.proto
        )
        self.proxy = self.proto.make_proxy(DemoInterface)

    async def expect_data(self, expected: bytes) -> None:
        data = await async_read_fd(self.pipers, len(expected) + 1)
        self.assertEqual(data, expected)

    async def test_broken_pipe_send(self) -> None:
        os.close(self.pipers)
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        with self.assertRaises(BrokenPipeError):
            await fut
        with self.assertRaisesRegex(OSError, "Bad file descriptor"):
            os.close(self.pipewc)

    async def test_broken_pipe_receive(self) -> None:
        fut = asyncio.ensure_future(self.proxy.Method(argument="spam"))
        await self.expect_data(
            b'{"method":"com.example.demo.Method","parameters":{"argument":"spam"}}\0'
        )
        self.assertFalse(fut.done())
        os.close(self.pipews)
        with self.assertRaises(ConnectionResetError):
            await fut
        with self.assertRaisesRegex(OSError, "Bad file descriptor"):
            os.close(self.pipewc)



#!/usr/bin/python3
# Copyright 2026 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import argparse
import asyncio
import dataclasses
import math
import secrets
import socket
import sys
import typing
import unittest

if __file__.split("/")[-2:-1] == ["tests"]:
    sys.path.insert(0, "/".join(__file__.split("/")[:-2]))

from asyncvarlink import (
    connect_unix_varlink,
    create_unix_server,
    TypedVarlinkErrorReply,
    VarlinkClientProtocol,
    VarlinkInterface,
    VarlinkInterfaceRegistry,
    varlinkmethod,
    VarlinkMethodCall,
    VarlinkTransport,
)
from asyncvarlink.serviceinterface import VarlinkServiceInterface


class ResultTest05(typing.TypedDict):
    bool: bool
    int: int
    float: float
    string: str


class ResultTest06(typing.TypedDict):
    struct: ResultTest05


class Anon(typing.TypedDict):
    foo: bool
    bar: bool


@dataclasses.dataclass
class Interface:
    foo: list[dict[str, typing.Literal["foo", "bar", "baz"]] | None] | None
    anon: Anon


class FirstSecondStruct(typing.TypedDict):
    first: int
    second: str


@dataclasses.dataclass
class MyType:
    object: object
    enum: typing.Literal["one", "two", "three"]
    struct: FirstSecondStruct
    array: list[str]
    dictionary: dict[str, str]
    stringset: set[str]
    nullable: str | None
    nullable_array_struct: list[FirstSecondStruct] | None
    interface: Interface


class ClientIdError(
    TypedVarlinkErrorReply, interface="org.varlink.certification"
):
    class Parameters:
        pass


class CertificationError(
    TypedVarlinkErrorReply, interface="org.varlink.certification"
):
    class Parameters:
        wants: str
        got: str


class CertificationInterface(
    VarlinkInterface, name="org.varlink.certification"
):
    errors = (ClientIdError, CertificationError)

    def __init__(self) -> None:
        self.next_method: dict[str, str] = {}

    def check_equal(self, wants: typing.Any, got: typing.Any) -> None:
        if wants != got:
            if isinstance(wants, str) and isinstance(got, str):
                raise CertificationError(wants=wants, got=got)
            raise CertificationError(wants=repr(wants), got=repr(got))

    def check_method(
        self, client_id: str, cur_meth: str, next_meth: str
    ) -> None:
        try:
            exp_meth = self.next_method[client_id]
        except KeyError:
            raise CertificationError(
                wants="org.varlink.certification.Start",
                got="org.varlink.certification." + cur_meth,
            ) from None
        self.check_equal(
            "org.varlink.certification." + exp_meth,
            "org.varlink.certification." + cur_meth,
        )
        self.next_method[client_id] = next_meth

    @varlinkmethod(return_parameter="client_id")
    def Start(self) -> str:
        client_id = secrets.token_hex()
        self.next_method[client_id] = "Test01"
        return client_id

    @varlinkmethod(return_parameter="bool")
    def Test01(self, client_id: str) -> bool:
        self.check_method(client_id, "Test01", "Test02")
        return True

    @varlinkmethod(return_parameter="int")
    def Test02(self, client_id: str, bool: bool) -> int:
        self.check_method(client_id, "Test02", "Test03")
        self.check_equal(True, bool)
        return 1

    @varlinkmethod(return_parameter="float")
    def Test03(self, client_id: str, int: int) -> float:
        self.check_method(client_id, "Test03", "Test04")
        self.check_equal(1, int)
        return 1.0

    @varlinkmethod(return_parameter="string")
    def Test04(self, client_id: str, float: float) -> str:
        self.check_method(client_id, "Test04", "Test05")
        self.check_equal(1.0, float)
        return "ping"

    @varlinkmethod
    def Test05(self, client_id: str, string: str) -> ResultTest05:
        self.check_method(client_id, "Test05", "Test06")
        self.check_equal("ping", string)
        return {
            "bool": False,
            "int": 2,
            "float": math.pi,
            "string": "a lot of string",
        }

    @varlinkmethod
    def Test06(
        self, client_id: str, bool: bool, int: int, float: float, string: str
    ) -> ResultTest06:
        self.check_method(client_id, "Test06", "Test07")
        self.check_equal(False, bool)
        self.check_equal(2, int)
        self.check_equal(math.pi, float)
        self.check_equal("a lot of string", string)
        return {
            "struct": {
                "bool": False,
                "int": 2,
                "float": math.pi,
                "string": "a lot of string",
            },
        }

    @varlinkmethod(return_parameter="map")
    def Test07(self, client_id: str, struct: ResultTest05) -> dict[str, str]:
        self.check_method(client_id, "Test07", "Test08")
        self.check_equal(
            {
                "bool": False,
                "int": 2,
                "float": math.pi,
                "string": "a lot of string",
            },
            struct,
        )
        return {"foo": "Foo", "bar": "Bar"}

    @varlinkmethod(return_parameter="set")
    def Test08(self, client_id: str, map: dict[str, str]) -> set[str]:
        self.check_method(client_id, "Test08", "Test09")
        self.check_equal({"foo": "Foo", "bar": "Bar"}, map)
        return {"one", "two", "three"}

    @varlinkmethod(return_parameter="mytype")
    def Test09(self, client_id: str, set: set[str]) -> MyType:
        self.check_method(client_id, "Test09", "Test10")
        self.check_equal({"one", "two", "three"}, set)
        return MyType(
            object={
                "method": "org.varlink.certification.Test09",
                "parameters": {"map": {"foo": "Foo", "bar": "Bar"}},
            },
            enum="two",
            struct={"first": 1, "second": "2"},
            array=["one", "two", "three"],
            dictionary={"foo": "Foo", "bar": "Bar"},
            stringset={"one", "two", "three"},
            nullable=None,
            nullable_array_struct=None,
            interface=Interface(
                foo=[
                    None,
                    {"foo": "foo", "bar": "bar"},
                    None,
                    {"one": "foo", "two": "bar"},
                ],
                anon={"foo": True, "bar": False},
            ),
        )

    @varlinkmethod(return_parameter="string")
    def Test10(self, client_id: str, mytype: MyType) -> typing.Iterator[str]:
        self.check_equal(
            {
                "method": "org.varlink.certification.Test09",
                "parameters": {"map": {"foo": "Foo", "bar": "Bar"}},
            },
            mytype.object,
        )
        self.check_equal("two", mytype.enum)
        self.check_equal({"first": 1, "second": "2"}, mytype.struct)
        self.check_equal(["one", "two", "three"], mytype.array)
        self.check_equal({"foo": "Foo", "bar": "Bar"}, mytype.dictionary)
        self.check_equal({"one", "two", "three"}, mytype.stringset)
        self.check_equal(None, mytype.nullable)
        self.check_equal(None, mytype.nullable_array_struct)
        self.check_equal(
            Interface(
                foo=[
                    None,
                    {"foo": "foo", "bar": "bar"},
                    None,
                    {"one": "foo", "two": "bar"},
                ],
                anon={"foo": True, "bar": False},
            ),
            mytype.interface,
        )
        self.check_method(client_id, "Test10", "Test11")
        for i in range(1, 11):
            yield f"Reply number {i}"

    @varlinkmethod
    def Test11(self, client_id: str, last_more_replies: list[str]) -> None:
        '''must be called as "oneway"'''
        self.check_method(client_id, "Test11", "End")
        self.check_equal(
            [f"Reply number {i}" for i in range(1, 11)], last_more_replies
        )

    @varlinkmethod(return_parameter="all_ok")
    def End(self, client_id: str) -> bool:
        self.check_method(client_id, "End", "Start")
        del self.next_method[client_id]
        return True


def make_registry() -> VarlinkInterfaceRegistry:
    registry = VarlinkInterfaceRegistry()
    registry.register_interface(
        VarlinkServiceInterface(
            "asyncvarlink",
            "certification_server",
            "0",
            "https://github.com/helmutg/asyncvarlink",
            registry,
        ),
    )
    registry.register_interface(CertificationInterface())
    return registry


async def run_test(protocol: VarlinkClientProtocol) -> None:
    proxy = protocol.make_proxy(CertificationInterface)
    res = await proxy.Start()
    cid = res["client_id"]
    res = await proxy.Test01(client_id=cid)
    res = await proxy.Test02(client_id=cid, **res)
    res = await proxy.Test03(client_id=cid, **res)
    res = await proxy.Test04(client_id=cid, **res)
    res = await proxy.Test05(client_id=cid, **res)
    res = await proxy.Test06(client_id=cid, **res)
    res = await proxy.Test07(client_id=cid, **res)
    res = await proxy.Test08(client_id=cid, **res)
    res = await proxy.Test09(client_id=cid, **res)
    array = []
    async for res in proxy.Test10(client_id=cid, **res):
        array.append(res["string"])
    await proxy.Test11.oneway(client_id=cid, last_more_replies=array)
    if not await proxy.End(client_id=cid):
        raise RuntimeError("certification failed")


async def selftest() -> None:
    loop = asyncio.get_running_loop()
    sock1, sock2 = socket.socketpair()
    sp = make_registry().protocol_factory()
    st = VarlinkTransport(loop, sock1, sock1, sp)
    cp = VarlinkClientProtocol()
    ct = VarlinkTransport(loop, sock2, sock2, cp)
    await asyncio.sleep(0)
    await run_test(cp)
    ct.close()
    st.close()


class CertificationTests(unittest.IsolatedAsyncioTestCase):
    async def test_self(self) -> None:
        await selftest()


async def client_main(sock: str) -> None:
    _, p = await connect_unix_varlink(VarlinkClientProtocol, sock)
    assert isinstance(p, VarlinkClientProtocol)
    await run_test(p)
    print("Certification passed")


async def server_main(sock: str) -> None:
    await (
        await create_unix_server(
            make_registry().protocol_factory,
            sock,
            loop=asyncio.get_running_loop(),
        )
    ).serve_forever()


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--client", action="store_true")
    parser.add_argument("--socket", action="store")
    args = parser.parse_args()
    if not args.socket:
        asyncio.run(selftest())
        print("Certification passed")
    elif args.client:
        asyncio.run(client_main(args.socket))
    else:
        asyncio.run(server_main(args.socket))


if __name__ == "__main__":
    main()



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import collections.abc
import typing
import unittest

from asyncvarlink.interface import (
    AnnotatedResult,
    LastResult,
    VarlinkInterface,
    varlinkmethod,
)


class ResultWrapper(typing.TypedDict):
    result: str


def res(val: str) -> ResultWrapper:
    return {"result": val}


class ExpectedError(Exception):
    pass


class TestInterface(unittest.TestCase):
    def test_synchronous(self) -> None:
        class SyncInterface(VarlinkInterface):
            name = "com.example.SyncInterface"

            def __init__(self) -> None:
                self.gen_state = -1
                self.genr_state = -1
                self.geni_state = -1
                self.gene_state = -1

            @varlinkmethod
            def simple(self) -> ResultWrapper:
                return res("simple")

            @varlinkmethod
            def annotated(self) -> ResultWrapper:
                return AnnotatedResult(res("annotated"))

            @varlinkmethod(return_parameter="result")
            def named(self) -> str:
                return "named"

            @varlinkmethod(return_parameter="result")
            def annotated_named(self) -> str:
                return AnnotatedResult(res("annotated_named"))

            @varlinkmethod(return_parameter="result")
            def iterable(self) -> list[int]:
                return []

            @varlinkmethod
            def error(self) -> ResultWrapper:
                raise ExpectedError()

            @varlinkmethod
            def gen(self) -> collections.abc.Iterator[ResultWrapper]:
                self.gen_state = 0
                yield AnnotatedResult(res("gen0"), continues=True)
                self.gen_state = 1
                yield res("gen1")
                self.gen_state = 2
                yield res("gen2")
                self.gen_state = 3

            @varlinkmethod(return_parameter="result")
            def gen_raise(self) -> collections.abc.Iterator[str]:
                self.genr_state = 0
                yield "genr0"
                self.genr_state = 1
                raise LastResult("genr1")

            @varlinkmethod(delay_generator=False, return_parameter="result")
            def gen_immediate(self) -> collections.abc.Iterator[str]:
                self.geni_state = 0
                yield AnnotatedResult(res("geni0"), continues=True)
                self.geni_state = 1
                yield "geni1"
                self.geni_state = 2
                raise LastResult("geni2")

            @varlinkmethod(return_parameter="result")
            def gen_error(self) -> collections.abc.Iterator[str]:
                self.gene_state = 0
                yield "gene0"
                self.gene_state = 1
                raise ExpectedError()

            @varlinkmethod
            def optional(self, *, optional: int | None = None) -> None:
                if optional is not None:
                    raise ValueError("unexpected optional value")

        iface = SyncInterface()
        self.assertEqual(iface.simple(), AnnotatedResult(res("simple")))
        self.assertEqual(iface.annotated(), AnnotatedResult(res("annotated")))
        self.assertEqual(iface.named(), AnnotatedResult(res("named")))
        self.assertEqual(
            iface.annotated_named(), AnnotatedResult(res("annotated_named"))
        )
        self.assertEqual(iface.iterable(), AnnotatedResult({"result": []}))
        self.assertRaises(ExpectedError, iface.error)

        it = iface.gen()
        self.assertEqual(iface.gen_state, -1)
        self.assertEqual(
            next(it), AnnotatedResult(res("gen0"), continues=True)
        )
        self.assertEqual(iface.gen_state, 0)
        self.assertEqual(
            next(it), AnnotatedResult(res("gen1"), continues=True)
        )
        self.assertGreater(iface.gen_state, 1)
        self.assertLess(iface.gen_state, 3)
        self.assertEqual(next(it), AnnotatedResult(res("gen2")))
        self.assertEqual(iface.gen_state, 3)
        self.assertRaises(StopIteration, next, it)

        it = iface.gen_raise()
        self.assertEqual(iface.genr_state, -1)
        self.assertEqual(
            next(it), AnnotatedResult(res("genr0"), continues=True)
        )
        self.assertEqual(iface.genr_state, 1)
        self.assertEqual(next(it), AnnotatedResult(res("genr1")))
        self.assertRaises(StopIteration, next, it)

        it = iface.gen_immediate()
        self.assertEqual(iface.geni_state, -1)
        self.assertEqual(
            next(it), AnnotatedResult(res("geni0"), continues=True)
        )
        self.assertEqual(iface.geni_state, 0)
        self.assertEqual(
            next(it), AnnotatedResult(res("geni1"), continues=True)
        )
        self.assertEqual(iface.geni_state, 1)
        self.assertEqual(next(it), AnnotatedResult(res("geni2")))
        self.assertEqual(iface.geni_state, 2)
        self.assertRaises(StopIteration, next, it)
        it = iface.gen_error()
        self.assertEqual(iface.gene_state, -1)
        self.assertEqual(
            next(it), AnnotatedResult(res("gene0"), continues=True)
        )
        self.assertEqual(iface.gene_state, 1)
        self.assertRaises(ExpectedError, next, it)
        iface.optional()


class TestAsyncInterface(unittest.IsolatedAsyncioTestCase):
    async def test_async(self) -> None:
        class AsyncInterface(VarlinkInterface):
            name = "com.example.AsyncInterface"

            def __init__(self) -> None:
                self.gen_state = -1
                self.genr_state = -1
                self.geni_state = -1

            @varlinkmethod
            async def simple(self) -> ResultWrapper:
                return res("simple")

            @varlinkmethod
            async def annotated(self) -> ResultWrapper:
                return AnnotatedResult(res("annotated"))

            @varlinkmethod(return_parameter="result")
            async def named(self) -> str:
                return "named"

            @varlinkmethod(return_parameter="result")
            async def annotated_named(self) -> str:
                return AnnotatedResult(res("annotated_named"))

            @varlinkmethod
            async def gen(
                self,
            ) -> collections.abc.AsyncIterator[ResultWrapper]:
                self.gen_state = 0
                yield AnnotatedResult(res("gen0"), continues=True)
                self.gen_state = 1
                yield res("gen1")
                self.gen_state = 2
                yield res("gen2")
                self.gen_state = 3

            @varlinkmethod(return_parameter="result")
            async def gen_raise(self) -> collections.abc.AsyncIterator[str]:
                self.genr_state = 0
                yield "genr0"
                self.genr_state = 1
                raise LastResult("genr1")

            @varlinkmethod(delay_generator=False, return_parameter="result")
            async def gen_immediate(
                self,
            ) -> collections.abc.AsyncIterator[str]:
                self.geni_state = 0
                yield AnnotatedResult(res("geni0"), continues=True)
                self.geni_state = 1
                yield "geni1"
                self.geni_state = 2
                raise LastResult("geni2")

            @varlinkmethod
            async def optional(self, *, optional: int | None = None) -> None:
                if optional is not None:
                    raise ValueError("unexpected optional value")

        iface = AsyncInterface()
        self.assertEqual(await iface.simple(), AnnotatedResult(res("simple")))
        self.assertEqual(
            await iface.annotated(), AnnotatedResult(res("annotated"))
        )
        self.assertEqual(await iface.named(), AnnotatedResult(res("named")))
        self.assertEqual(
            await iface.annotated_named(),
            AnnotatedResult(res("annotated_named")),
        )

        it = iface.gen()
        self.assertEqual(iface.gen_state, -1)
        self.assertEqual(
            await anext(it), AnnotatedResult(res("gen0"), continues=True)
        )
        self.assertEqual(iface.gen_state, 0)
        self.assertEqual(
            await anext(it), AnnotatedResult(res("gen1"), continues=True)
        )
        self.assertGreater(iface.gen_state, 1)
        self.assertLess(iface.gen_state, 3)
        self.assertEqual(await anext(it), AnnotatedResult(res("gen2")))
        self.assertEqual(iface.gen_state, 3)
        with self.assertRaises(StopAsyncIteration):
            await anext(it)

        it = iface.gen_raise()
        self.assertEqual(iface.genr_state, -1)
        self.assertEqual(
            await anext(it), AnnotatedResult(res("genr0"), continues=True)
        )
        self.assertEqual(iface.genr_state, 1)
        self.assertEqual(await anext(it), AnnotatedResult(res("genr1")))
        with self.assertRaises(StopAsyncIteration):
            await anext(it)

        it = iface.gen_immediate()
        self.assertEqual(iface.geni_state, -1)
        self.assertEqual(
            await anext(it), AnnotatedResult(res("geni0"), continues=True)
        )
        self.assertEqual(iface.geni_state, 0)
        self.assertEqual(
            await anext(it), AnnotatedResult(res("geni1"), continues=True)
        )
        self.assertEqual(iface.geni_state, 1)
        self.assertEqual(await anext(it), AnnotatedResult(res("geni2")))
        self.assertEqual(iface.geni_state, 2)
        with self.assertRaises(StopAsyncIteration):
            await anext(it)
        await iface.optional()



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""asyncio and protocol level varlink functionality."""

import asyncio
import collections
import collections.abc
import errno
import functools
import json
import logging
import os
import socket
import typing
import warnings

from .types import (
    FileDescriptor,
    FileDescriptorArray,
    HasFileno,
    HasFilenoAndClose,
    JSONObject,
    override,
)


_logger = logging.getLogger("asyncvarlink.protocol")


_SENTINEL = object()


class VarlinkBaseProtocol(asyncio.BaseProtocol):
    """An asyncio protocol that provides the interface expected by
    VarlinkTransport, but does not actually implement any of the wire protocol.
    The typical data_received used by other streaming protocols is replaced
    with message_received.
    """

    def message_received(
        self, data: bytes, fds: FileDescriptorArray | None
    ) -> None:
        """Called when the transport received new data. The data can be
        accompanied by open file descriptors. The caller will release the fds
        array and all contained file descriptors that have not been taken
        (FileDescriptorArray.take) once message_received returns None. Life
        time of fds can be extended by adding a reference prior to returning.
        """

    def eof_received(self) -> None:
        """Callback for signalling the end of messages on the receiving side.
        The default implementation does nothing. Same as Protocol.eof_received.
        """


def _check_socket(thing: socket.socket | int | HasFileno) -> HasFilenoAndClose:
    """Attempt to upgrade a file descriptor into a socket object if it happens
    to be a socket.
    """
    if isinstance(thing, socket.socket):
        return thing
    if not hasattr(thing, "fileno"):
        if not isinstance(thing, int):
            raise TypeError("not a file descriptor")
        thing = FileDescriptor(thing)
    elif not hasattr(thing, "close"):
        thing = FileDescriptor(thing)
    elif typing.TYPE_CHECKING:
        # In principle, it could be an incompatible close method, so mypy
        # rejects this otherwise.
        thing = typing.cast(HasFilenoAndClose, thing)
    try:
        sock = socket.socket(fileno=thing.fileno())
    except OSError as err:
        if err.errno == errno.ENOTSOCK:
            return thing
        raise
    if sock.type != socket.SOCK_STREAM:
        raise ValueError("the given socket is not SOCK_STREAM")
    return sock


_BLOCKING_ERRNOS = frozenset((errno.EWOULDBLOCK, errno.EAGAIN))


# pylint: disable=too-many-instance-attributes  # Yes, we need them all
class VarlinkTransport(asyncio.BaseTransport):
    """A specialized asyncio Transport class for use with varlink and file
    descriptor passing. As such, it provides send_message rather than write
    and expects the protocol class to provide message_received rather than
    read. It also allows sending and receiving to happen on different file
    descriptors to facilitate use with stdin and stdout pipes.
    """

    MAX_RECV_FDS = 1024
    """The maximum number of file descriptors that will be expected in a single
    varlink message.
    """

    closed_future: asyncio.Future[None]
    """A future that becomes ready once the transport is closed."""

    # pylint: disable=too-many-arguments  # Yes, we need five arguments.
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        recvfd: socket.socket | int | HasFileno,
        sendfd: socket.socket | int | HasFileno,
        protocol: VarlinkBaseProtocol,
        extra: collections.abc.Mapping[str, typing.Any] | None = None,
    ):
        super().__init__(extra)
        self._loop = loop
        self._recvfd: HasFilenoAndClose | None = _check_socket(recvfd)
        self._paused = True
        os.set_blocking(self._recvfd.fileno(), False)
        self._sendfd: HasFilenoAndClose | None
        if recvfd is sendfd:
            self._sendfd = self._recvfd
        else:
            self._sendfd = _check_socket(sendfd)
            os.set_blocking(self._sendfd.fileno(), False)
        if isinstance(self._sendfd, socket.socket):
            self._do_write = self._do_write_socket
        else:
            self._do_write = self._do_write_fd
        # Using a deque as we only do end operations and those become O(1).
        self._sendqueue: collections.deque[
            tuple[
                list[bytes],  # data to be sent
                list[int],  # file descriptors to be sent
                asyncio.Future[None],  # completion notification
            ]
        ]
        self._sendqueue = collections.deque()
        self._closing = False
        self.set_protocol(protocol)
        self._loop.call_soon(self._protocol.connection_made, self)
        self._loop.call_soon(self.resume_reading)
        self.closed_future = self._loop.create_future()

    @override
    def set_protocol(self, protocol: asyncio.BaseProtocol) -> None:
        assert isinstance(protocol, VarlinkBaseProtocol)
        self._protocol = protocol

    @override
    def get_protocol(self) -> VarlinkBaseProtocol:
        return self._protocol

    def _close_receiver(self) -> None:
        if self._recvfd is None:
            return
        loose_connection = False
        if self._sendfd is None:
            loose_connection = not self._closing
            self._closing = True
        if not self._paused:
            self._loop.remove_reader(self._recvfd)
        if (
            self._sendfd is None
            or self._recvfd.fileno() != self._sendfd.fileno()
        ):
            self._recvfd.close()
        self._recvfd = None
        if loose_connection:
            self._loop.call_soon(self._connection_lost)

    def _handle_read_socket(self) -> None:
        assert isinstance(self._recvfd, socket.socket)
        try:
            msg, fds, _flags, _addr = socket.recv_fds(
                self._recvfd, 4096, self.MAX_RECV_FDS
            )
        except OSError as err:
            if err.errno in _BLOCKING_ERRNOS:
                return
            _logger.debug(
                "%r: reading from socket failed", self, exc_info=True
            )
            self._loop.remove_reader(self._recvfd)
            self._close_receiver()
            return
        if msg:
            ownedfds = FileDescriptorArray(_SENTINEL, fds) if fds else None
            try:
                self._protocol.message_received(msg, ownedfds)
            finally:
                if ownedfds:
                    ownedfds.release(_SENTINEL)
        else:
            for fd in fds:
                os.close(fd)
            self._loop.remove_reader(self._recvfd)
            try:
                self._protocol.eof_received()
            finally:
                self._close_receiver()

    def _handle_read_fd(self) -> None:
        assert self._recvfd is not None
        try:
            data = os.read(self._recvfd.fileno(), 4096)
        except OSError as err:
            if err.errno in _BLOCKING_ERRNOS:
                return
            _logger.debug(
                "%r: reading from socket failed", self, exc_info=True
            )
            self._loop.remove_reader(self._recvfd)
            self._close_receiver()
            return
        if data:
            self._protocol.message_received(data, None)
        else:
            self._loop.remove_reader(self._recvfd)
            try:
                self._protocol.eof_received()
            finally:
                self._close_receiver()

    def pause_reading(self) -> None:
        """Pause receiving messages. No data will be passed to the protocol's
        message_received() method until resume_reading is called.
        """
        if self._closing or self._recvfd is None or self._paused:
            return
        self._paused = True
        self._loop.remove_reader(self._recvfd)

    def pause_receiving(self) -> None:
        """Deprecated alias of pause_reading."""
        warnings.warn(
            "method renamed to pause_reading",
            DeprecationWarning,
            stacklevel=2,
        )
        self.pause_reading()

    def resume_reading(self) -> None:
        """Resume receiving messages. Received messages will be passed to the
        protocol's message_received method again.
        """
        if self._closing or self._recvfd is None or not self._paused:
            return
        self._paused = False
        if isinstance(self._recvfd, socket.socket):
            self._loop.call_soon(
                self._loop.add_reader, self._recvfd, self._handle_read_socket
            )
        else:
            self._loop.call_soon(
                self._loop.add_reader, self._recvfd, self._handle_read_fd
            )

    def resume_receiving(self) -> None:
        """Deprecated alias of resume_reading."""
        warnings.warn(
            "method renamed to resume_reading",
            DeprecationWarning,
            stacklevel=2,
        )
        self.resume_reading()

    def send_message(
        self, data: bytes, fds: list[int] | None = None
    ) -> asyncio.Future[None]:
        """Enqueue the given data and file descriptors for sending. In case
        file descriptors are provided, they will be delivered combined using
        sendmsg. Otherwise, messages may be concatenated. The returned future
        will be done when the message has been sent. The given file descriptors
        should remain open until then and the responsibility for closing them
        remains with the caller.
        """
        if self._do_write is self._do_write_fd and fds:
            raise ValueError("cannot send fds on non-socket transport")
        if fds is None:
            fds = []
        if self._closing or self._sendfd is None:
            _logger.warning("%r: attempt to write to closed transport", self)
            fut = self._loop.create_future()
            fut.set_exception(BrokenPipeError())
            return fut
        if self._sendqueue:
            lastitem = self._sendqueue[-1]
            if lastitem[1] or fds:
                fut = self._loop.create_future()
                self._sendqueue.append(([data], fds, fut))
            else:
                fut = lastitem[2]
                lastitem[0].append(data)
        else:
            fut = self._loop.create_future()
            self._sendqueue.append(([data], fds, fut))
            self._loop.call_soon(
                self._loop.add_writer, self._sendfd, self._handle_write
            )
        return fut

    def _fail_sendqueue(self) -> None:
        while self._sendqueue:
            _, _, fut = self._sendqueue.popleft()
            fut.set_exception(BrokenPipeError())

    def _close_sender(self) -> None:
        if self._sendfd is None:
            return
        self._loop.remove_writer(self._sendfd)
        loose_connection = False
        if self._recvfd is None:
            loose_connection = not self._closing
            self._closing = True
        if (
            self._recvfd is None
            or self._recvfd.fileno() != self._sendfd.fileno()
        ):
            self._sendfd.close()
        self._sendfd = None
        if loose_connection:
            self._loop.call_soon(self._connection_lost)

    def _handle_write(self) -> None:
        assert self._sendfd is not None
        while self._sendqueue:
            data, fds, fut = self._sendqueue.popleft()
            try:
                sent = self._do_write(data, fds)
            except OSError as err:
                if err.errno in _BLOCKING_ERRNOS:
                    self._sendqueue.appendleft((data, fds, fut))
                else:
                    _logger.debug("%r: sending failed", self, exc_info=True)
                    self._close_sender()
                    fut.set_exception(err)
                    self._fail_sendqueue()
                return
            while sent > 0:
                assert data
                if sent >= len(data[0]):
                    sent -= len(data.pop(0))
                else:
                    data[0] = data[0][:sent]
                    sent = 0
            if data:
                self._sendqueue.appendleft((data, [], fut))
            else:
                fut.set_result(None)
        if not self._sendqueue:
            if self._closing:
                self._close_sender()
            else:
                self._loop.remove_writer(self._sendfd)

    def _do_write_socket(self, data: list[bytes], fds: list[int]) -> int:
        assert isinstance(self._sendfd, socket.socket)
        if fds:
            return socket.send_fds(self._sendfd, data, fds)
        return self._sendfd.sendmsg(data)

    def _do_write_fd(self, data: list[bytes], fds: list[int]) -> int:
        assert not fds
        assert self._sendfd is not None
        return os.writev(self._sendfd.fileno(), data)

    def _connection_lost(self) -> None:
        try:
            self._protocol.connection_lost(None)
        finally:
            self._close_receiver()
            if not self._sendqueue:
                self._close_sender()
            self.closed_future.set_result(None)

    @override
    def close(self) -> None:
        if not self._closing:
            self._closing = True
            self._loop.call_soon(self._connection_lost)

    @override
    def is_closing(self) -> bool:
        return self._closing

    def abort(self) -> None:
        """Close the transport immediately.

        In addition to closing, queued transfers will be cancelled.
        """
        self.close()
        self._fail_sendqueue()
        # Will soon call self._connection_lost, which immediately closes both
        # fds as the _sendqueue is now empty.


_JSONEncoder = json.JSONEncoder(separators=(",", ":"))


class VarlinkProtocol(VarlinkBaseProtocol):
    """An asyncio protocol that provides message_received() rather than
    data_received() to accommodate passed file descriptors.
    """

    def __init__(self) -> None:
        self._recv_buffer = b""
        self._consumer_queue: collections.deque[
            tuple[
                # A closure for invoking the next consumer
                collections.abc.Callable[[], asyncio.Future[None] | None],
                # An optional future that should be notified when the consumer
                # is done consuming (i.e. the future it returned is completed
                # or it returned None or raised an exception).
                asyncio.Future[None] | None,
            ]
        ] = collections.deque()
        self._transport: VarlinkTransport | None = None

    @override
    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        assert isinstance(transport, VarlinkTransport)
        self._transport = transport

    @override
    def message_received(
        self, data: bytes, fds: FileDescriptorArray | None
    ) -> None:
        """Handle incoming data by parsing the low-level part of the varlink
        protocol. Pass on received data via request_received.
        """
        parts = data.split(b"\0")
        if self._recv_buffer:
            parts[0] = self._recv_buffer + parts[0]
        self._recv_buffer = parts.pop()
        loop = asyncio.get_running_loop()
        processing = bool(self._consumer_queue)
        for reqdata in parts:
            fut = None
            if fds:
                fut = loop.create_future()
                fds.reference_until_done(fut)
            call: collections.abc.Callable[[], asyncio.Future[None] | None]
            try:
                obj = json.loads(reqdata)
            except json.decoder.JSONDecodeError as err:
                call = functools.partial(
                    self.error_received, err, reqdata, fds
                )
            else:
                if isinstance(obj, dict):
                    call = functools.partial(self.request_received, obj, fds)
                else:
                    call = functools.partial(
                        self.error_received,
                        TypeError("received JSON value is not an object"),
                        reqdata,
                        fds,
                    )
            self._consumer_queue.append((call, fut))
            if not processing:
                loop.call_soon(self._process_queue, None)
                processing = True
            fds = None

    def _process_queue(self, fut: asyncio.Future[None] | None) -> None:
        if fut is not None:
            assert fut.done()
            exc = fut.exception()
            if exc is not None:
                _logger.error(
                    "unhandled exception in future from request_received",
                    exc_info=exc,
                )
            fut = None
        if not self._consumer_queue:
            if self._transport is not None:
                self._transport.resume_reading()
            return
        consume, notify = self._consumer_queue.popleft()
        try:
            fut = consume()
        except Exception as exception:
            _logger.error(
                "unhandled exception from request_received or error_received",
                exc_info=exception,
            )
            raise
        finally:
            if notify is not None:
                notify.set_result(None)
            if fut is not None and fut.done():
                exc = fut.exception()
                if exc is not None:
                    _logger.error(
                        "unhandled exception in future from request_received",
                        exc_info=exc,
                    )
                fut = None
            if fut is None:
                # If the consumer finishes immediately, skip back pressure
                # via pause_reading as that typically incurs two syscalls.
                if self._consumer_queue:
                    asyncio.get_running_loop().call_soon(
                        self._process_queue, None
                    )
                elif self._transport is not None:
                    self._transport.resume_reading()
            else:
                fut.add_done_callback(self._process_queue)
                if self._transport is not None:
                    self._transport.pause_reading()

    def request_received(
        self, obj: JSONObject, fds: FileDescriptorArray | None
    ) -> asyncio.Future[None] | None:
        """Handle an incoming varlink request or response object together with
        associated file descriptors. If the handler returns a future, further
        processing will be delayed until the future is done. Once the function
        returns None, the fds will be released unless its life time has been
        extended by adding another referee.
        """
        raise NotImplementedError

    # pylint: disable=unused-argument  # Arguments provided for inheritance
    def error_received(
        self, err: Exception, data: bytes, fds: FileDescriptorArray | None
    ) -> None:
        """Handle an incoming protocol violation such as wrongly encoded JSON.
        The default handler does nothing.
        """

    def send_message(
        self,
        obj: JSONObject,
        fds: list[int] | None = None,
        autoclose: bool = True,
    ) -> asyncio.Future[None]:
        """Send a varlink request or response together with associated file
        descriptors. The returned future is done once the message actually has
        been sent or terminally failed sending. In the latter case, the future
        raises an exception. If autoclose is True, the file descriptors are
        closed after transmission. Otherwise, the caller is responsible for
        closing them after completion of the returned future.
        """
        if self._transport is None:
            fut = asyncio.get_running_loop().create_future()
            fut.set_exception(BrokenPipeError())
            return fut
        fut = self._transport.send_message(
            _JSONEncoder.encode(obj).encode("utf8") + b"\0", fds
        )
        if fds and autoclose:

            @fut.add_done_callback
            def close_fds(_: asyncio.Future[None]) -> None:
                for fd in fds:
                    os.close(fd)

        return fut

    def close(self) -> None:
        """Close the connected transport if any."""
        if self._transport is not None:
            try:
                self._transport.close()
            finally:
                self._transport = None



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import unittest

from asyncvarlink import AnnotatedResult, VarlinkInterfaceRegistry, override
from asyncvarlink.serviceinterface import (
    InterfaceNotFound,
    VarlinkServiceInterface,
)


class TestServiceInterface(unittest.TestCase):
    @override
    def setUp(self) -> None:
        self.reg = VarlinkInterfaceRegistry()
        self.vsi = VarlinkServiceInterface(
            "examplevendor",
            "exampleproduct",
            "1.0",
            "https://example.com",
            self.reg,
            comments=False,
        )
        self.reg.register_interface(self.vsi)

    def test_getinfo(self) -> None:
        self.assertEqual(self.vsi.name, "org.varlink.service")
        self.assertEqual(
            AnnotatedResult(
                {
                    "vendor": "examplevendor",
                    "product": "exampleproduct",
                    "version": "1.0",
                    "url": "https://example.com",
                    "interfaces": [self.vsi.name],
                },
            ),
            self.vsi.GetInfo(),
        )

    def test_description(self) -> None:
        self.assertEqual(
            AnnotatedResult(
                {
                    "description": """interface org.varlink.service

method GetInfo() -> (vendor: string, product: string, version: string, url: string, interfaces: []string)
method GetInterfaceDescription(interface: string) -> (description: string)

error InterfaceNotFound (interface: string)
error MethodNotFound (method: string)
error MethodNotImplemented (method: string)
error InvalidParameter (parameter: string)
error PermissionDenied ()
error ExpectedMore ()
""",
                },
            ),
            self.vsi.GetInterfaceDescription(interface=self.vsi.name),
        )

    def test_missing_interface(self) -> None:
        self.assertRaises(
            InterfaceNotFound,
            self.vsi.GetInterfaceDescription,
            interface="com.example.nonexistent",
        )



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Python classes for the org.varlink.service interface."""

import typing

from .interface import VarlinkInterface, varlinkmethod
from .serverprotocol import VarlinkInterfaceRegistry

# The serviceerrors module is split off to avoid circular imports, users should
# import its exceptions from this module.
from .serviceerrors import (
    ExpectedMore,
    InterfaceNotFound,
    InvalidParameter,
    MethodNotFound,
    MethodNotImplemented,
    PermissionDenied,
)


__all__ = [
    "ExpectedMore",
    "InterfaceNotFound",
    "InvalidParameter",
    "MethodNotFound",
    "MethodNotImplemented",
    "PermissionDenied",
    "VarlinkServiceInterface",
]


class VarlinkServiceInterface(VarlinkInterface, name="org.varlink.service"):
    """The Varlink Service Interface is provided by every varlink service. It
    describes the service and the interfaces it implements.
    """

    errors = (
        InterfaceNotFound,
        MethodNotFound,
        MethodNotImplemented,
        InvalidParameter,
        PermissionDenied,
        ExpectedMore,
    )

    class _GetInfoResult(typing.TypedDict):
        vendor: str
        product: str
        version: str
        url: str
        interfaces: list[str]

    def __init__(
        self,
        vendor: str,
        product: str,
        version: str,
        url: str,
        registry: VarlinkInterfaceRegistry,
        *,
        comments: bool = True,
    ):
        """Construct an introspection interface object from the given
        metadata and a VarlinkInterfaceRegistry for introspection. If
        comments is enabled (default), __doc__ strings will be included in
        interface descriptions as comments.
        """
        self._info: VarlinkServiceInterface._GetInfoResult = {
            "vendor": vendor,
            "product": product,
            "version": version,
            "url": url,
            "interfaces": [],
        }
        self._registry = registry
        self._comments = comments

    @varlinkmethod
    def GetInfo(self) -> _GetInfoResult:
        """Get a list of all the interfaces a service provides and information
        about the implementation.
        """
        return self._info | {
            "interfaces": sorted(iface.name for iface in self._registry)
        }

    @varlinkmethod(return_parameter="description")
    def GetInterfaceDescription(self, *, interface: str) -> str:
        """Get the description of an interface that is implemented by this
        service.
        """
        try:
            iface = self._registry[interface]
        except KeyError:
            raise InterfaceNotFound(interface=interface) from None
        return iface.render_interface_description(comments=self._comments)



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""asyncio varlink client protocol implementation"""

import asyncio
import collections
import collections.abc
import contextlib
import functools
import typing

from .conversion import FileDescriptorVarlinkType
from .interface import (
    VarlinkInterface,
    VarlinkMethodSignature,
    varlinksignature,
)
from .message import VarlinkMethodCall, VarlinkMethodReply
from .protocol import VarlinkProtocol
from .types import FileDescriptorArray, JSONObject, override
from .util import completing_future


class VarlinkClientProtocol(VarlinkProtocol):
    """Protocol class for a varlink client. Calls can be issued using the call
    and call_more methods. Associated replies are returned or generated from
    these methods asynchronously. There usually is no need to further derive
    this class.
    """

    _CallResult: typing.TypeAlias = tuple[
        VarlinkMethodReply, FileDescriptorArray | None
    ]

    class _PendingReply:
        def __init__(
            self,
            want_fds: bool,
            consumer_fut: asyncio.Future[None] | None = None,
        ):
            loop = asyncio.get_running_loop()
            self.future: asyncio.Future[
                "VarlinkClientProtocol._CallResult"
            ] = loop.create_future()
            self.want_fds = want_fds
            self.consumer_fut = consumer_fut or loop.create_future()

        def request_received(
            self, obj: JSONObject, fds: FileDescriptorArray | None
        ) -> None:
            if self.future.done():
                # Multiple replies not permitted.
                return
            try:
                reply = VarlinkMethodReply.fromjson(obj)
            except (TypeError, ValueError) as err:
                self.future.set_exception(err)
                return
            if not self.want_fds:
                fds = None
            elif fds:
                fds.reference_until_done(self.consumer_fut)
            self.future.set_result((reply, fds))

        def connection_lost(self, exc: Exception) -> None:
            if self.future.done():
                return
            self.future.set_exception(exc)

        def consumer_done(self) -> None:
            self.consumer_fut.set_result(None)

    class _PendingReplies:
        def __init__(self) -> None:
            self.replies = collections.deque(
                [VarlinkClientProtocol._PendingReply(True)]
            )

        def request_received(
            self, obj: JSONObject, fds: FileDescriptorArray | None
        ) -> None:
            lastreply = self.replies[-1]
            if lastreply.future.done():
                # This call does not handle any more continuations.
                return
            if obj.get("continues", False):
                self.replies.append(VarlinkClientProtocol._PendingReply(True))
            lastreply.request_received(obj, fds)

        def connection_lost(self, exc: Exception) -> None:
            lastreply = self.replies[-1]
            lastreply.connection_lost(exc)

        def consumer_done(self) -> None:
            if self.replies:
                while len(self.replies) > 1:
                    preply = self.replies.popleft()
                    assert preply.future.done()
                    preply.consumer_done()
                self.replies[0].consumer_done()

    def __init__(self) -> None:
        super().__init__()
        self._pending: collections.deque[
            VarlinkClientProtocol._PendingReply
            | VarlinkClientProtocol._PendingReplies
        ] = collections.deque()

    async def call(
        self,
        call: VarlinkMethodCall,
        fds: list[int] | None = None,
        replyfdsdone: asyncio.Future[None] | None = None,
    ) -> _CallResult | None:
        """Issue a varlink call. If the call has the more attribute set, the
        call_more method must be used instead. The given fds (if any) must
        remain available until the call method returns and the responsibility
        for closing them remains with the caller. If the caller expects file
        descriptors to be returned, it must pass a replyfdsdone future. The
        returned FileDescriptorArray (if any) will be valid until the
        replyfdsdone future is done.
        """
        assert not call.more
        if call.oneway:
            await self.send_message(call.tojson(), fds, autoclose=False)
            return None
        pending = VarlinkClientProtocol._PendingReply(
            replyfdsdone is not None, replyfdsdone
        )
        try:
            self._pending.append(pending)
            # We may pipeline calls here.
            await self.send_message(call.tojson(), fds, autoclose=False)
            return await pending.future
        finally:
            if replyfdsdone is None:
                pending.consumer_done()

    async def call_more(
        self, call: VarlinkMethodCall, fds: list[int] | None = None
    ) -> collections.abc.AsyncGenerator[_CallResult, None]:
        """Issue a varlink call expecting multiple replies. If the call does
        not set the more attribute, the call method must be used instead. The
        given fds (if any) must remain available until the method returns and
        the responsibility for closing them remains with the caller. Replies
        generally include returned a FileDescriptorArray if file descriptors
        were received. Such an array remains valid from one iteration to the
        next. Its lifetime can be extended using its reference_until_done
        method.
        """
        assert call.more
        pending = VarlinkClientProtocol._PendingReplies()
        self._pending.append(pending)
        # We may pipeline calls here.
        sendfut = self.send_message(call.tojson(), fds, autoclose=False)
        try:
            while pending.replies:
                preply = pending.replies[0]
                result = await preply.future
                pending.replies.popleft()
                try:
                    yield result
                finally:
                    preply.consumer_done()
        finally:
            pending.consumer_done()
            await sendfut

    @override
    def request_received(
        self, obj: JSONObject, fds: FileDescriptorArray | None
    ) -> None:
        if obj.get("continues", False):
            pending = self._pending[0]
        else:
            pending = self._pending.popleft()
        pending.request_received(obj, fds)

    @override
    def error_received(
        self,
        err: Exception,
        _data: bytes,
        _fds: FileDescriptorArray | None,
    ) -> None:
        self.connection_lost(err)

    @override
    def connection_lost(self, exc: Exception | None) -> None:
        super().connection_lost(exc)
        if exc is None:
            exc = ConnectionResetError("Connection lost")
        while self._pending:
            pending = self._pending.popleft()
            pending.connection_lost(exc)
        self.close()

    @override
    def eof_received(self) -> None:
        super().eof_received()
        self.connection_lost(ConnectionResetError("remote closed connection"))

    def make_proxy(
        self, interface: type[VarlinkInterface]
    ) -> "VarlinkInterfaceProxy":
        """Convenience function that constructs an interface proxy object for
        the combination of protocol instance and interface class at hand.
        """
        return VarlinkInterfaceProxy(self, interface)


class _ResourceManager:
    """A context manager that yields a given result on entry and calls a given
    cleanup function on exit.
    """

    def __init__(
        self, result: typing.Any, close: collections.abc.Callable[[], None]
    ):
        self._result = result
        self._close = close

    def __enter__(self) -> typing.Any:
        return self._result

    def __exit__(self, *_: object) -> None:
        self._close()


class _ProxiedVarlinkMethodBase:
    def __init__(
        self,
        protocol: VarlinkClientProtocol,
        interface: type[VarlinkInterface],
        fqmethod: str,
        signature: VarlinkMethodSignature,
    ):
        self._protocol = protocol
        self._interface = interface
        self._fqmethod = fqmethod
        self._signature = signature

    @property
    def oneway(self) -> "_ProxiedVarlinkOnewayMethod":
        return _ProxiedVarlinkOnewayMethod(
            self._protocol, self._interface, self._fqmethod, self._signature
        )

    def __call__(self, **kwargs: typing.Any) -> typing.Any:
        # This really returns either an Awaitable or an AsyncGenerator, but
        # typing it like that makes using it very annoying.
        raise NotImplementedError


class _ProxiedVarlinkOnewayMethod(_ProxiedVarlinkMethodBase):
    @property
    def oneway(self) -> typing.Self:
        return self

    @override
    async def __call__(self, **kwargs: typing.Any) -> None:
        with FileDescriptorArray.new_managed() as pfds:
            # This may raise a ConversionError
            parameters = self._signature.parameter_type.tojson(
                kwargs, {FileDescriptorVarlinkType: pfds}
            )
            with completing_future() as donefut:
                result = await self._protocol.call(
                    VarlinkMethodCall(self._fqmethod, parameters, oneway=True),
                    [fd.fileno() for fd in pfds],
                    donefut,
                )
                assert result is None


class _ProxiedVarlinkMoreMethod(_ProxiedVarlinkMethodBase):
    @override
    async def __call__(
        self, **kwargs: typing.Any
    ) -> collections.abc.AsyncGenerator[typing.Any, None]:
        with FileDescriptorArray.new_managed() as pfds:
            # This may raise a ConversionError.
            parameters = self._signature.parameter_type.tojson(
                kwargs, {FileDescriptorVarlinkType: pfds}
            )
            async for reply, rfds in self._protocol.call_more(
                VarlinkMethodCall(self._fqmethod, parameters, more=True),
                [fd.fileno() for fd in pfds],
            ):
                if reply.error is None:
                    # This may raise a ConversionError.
                    ret = self._signature.return_type.fromjson(
                        reply.parameters,
                        {FileDescriptorVarlinkType: rfds},
                    )
                    if self._signature.return_type.contains_fds:
                        if rfds is None:
                            yield contextlib.nullcontext(ret)
                        else:
                            sentinel = object()
                            rfds.reference(sentinel)
                            yield _ResourceManager(
                                ret,
                                functools.partial(rfds.release, sentinel),
                            )
                    else:
                        yield ret
                else:
                    self._interface.raise_error(reply)


class _ProxiedVarlinkMethod(_ProxiedVarlinkMethodBase):
    @override
    async def __call__(self, **kwargs: typing.Any) -> typing.Any:
        with FileDescriptorArray.new_managed() as pfds:
            # This may raise a ConversionError
            parameters = self._signature.parameter_type.tojson(
                kwargs, {FileDescriptorVarlinkType: pfds}
            )
            with completing_future() as donefut:
                result = await self._protocol.call(
                    VarlinkMethodCall(self._fqmethod, parameters),
                    [fd.fileno() for fd in pfds],
                    donefut,
                )
                assert result is not None
                reply, fda = result
                if reply.error is None:
                    # This may raise a ConversionError.
                    ret = self._signature.return_type.fromjson(
                        reply.parameters, {FileDescriptorVarlinkType: fda}
                    )
                    if not self._signature.return_type.contains_fds:
                        return ret
                    if fda is None:
                        return contextlib.nullcontext(ret)
                    sentinel = object()
                    fda.reference(sentinel)
                    return _ResourceManager(
                        ret, functools.partial(fda.release, sentinel)
                    )
                self._interface.raise_error(reply)


class VarlinkInterfaceProxy:
    """Interface proxy class for a varlink client. Its only purpose is to
    provide a suitable __getattr__ method such that methods declared on a
    VarlinkInterface subclass can be invoked on the proxy object and result in
    calls on a protocol instance.
    """

    def __init__(
        self,
        protocol: VarlinkClientProtocol,
        interface: type[VarlinkInterface],
    ):
        """The given protocol instance is used to actually perform calls. A
        protocol instance can be used with multiple proxy objects concurrently.
        The interface subclass (not instance) is used to generate conversions
        for proxy methods. Its methods can be stubbed as they are never called
        from a proxy. What matters is the types of its methods as they are used
        for converting between Python and JSON objects.
        """
        self._protocol = protocol
        self._interface = interface

    def __getattr__(self, attr: str) -> _ProxiedVarlinkMethodBase:
        """Look up a method on the VarlinkInterface subclass and return a proxy
        method if available. Raises AttributeError if no suitable method could
        be found. The type of the proxy method is dynamic. It is generally
        asynchronous and must be called within an asyncio event loop. It may or
        may not return a generator. All arguments must be keyword arguments.
        If the return type contains file descriptors, the result is
        additionally wrapped in a synchronous context manager.
        """
        try:
            fqmethod = f"{self._interface.name}.{attr}"
        except ValueError as err:
            raise AttributeError(f"invalid method name {attr}") from err
        # Intentionally propagate AttributeError
        method = getattr(self._interface, attr)
        if (signature := varlinksignature(method)) is None:
            raise AttributeError(f"no {attr} varlink method found")
        if signature.more:
            return _ProxiedVarlinkMoreMethod(
                self._protocol, self._interface, fqmethod, signature
            )
        return _ProxiedVarlinkMethod(
            self._protocol, self._interface, fqmethod, signature
        )



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Representation of varlink method calls and replies as Python objects."""

import dataclasses

from .types import JSONObject, JSONValue, validate_interface, validate_name


@dataclasses.dataclass
class VarlinkMethodCall:
    """Represent a parsed and roughly validated varlink method call."""

    method: str
    parameters: JSONObject
    oneway: bool = False
    more: bool = False
    upgrade: bool = False
    extensions: JSONObject = dataclasses.field(default_factory=dict)

    @property
    def method_interface(self) -> str:
        """Return the interface portion of the method string. May raise a
        ValueError.
        """
        interface, dot, _ = self.method.rpartition(".")
        if dot != ".":
            raise ValueError("unqualified method string")
        return interface

    @property
    def method_name(self) -> str:
        """Return the unqualified name portion of the method string."""
        return self.method.rpartition(".")[2]

    def __post_init__(self) -> None:
        validate_interface(self.method_interface)
        validate_name(self.method_name)

    @classmethod
    def fromjson(cls, obj: JSONValue) -> "VarlinkMethodCall":
        """Parse a JSON value into a validated VarlinkMethodCall. May raise
        TypeError and ValueError.
        """
        if not isinstance(obj, dict):
            raise TypeError(
                f"call object must be a map, is {obj.__class__.__name__}"
            )
        extensions = obj.copy()
        try:
            method = extensions.pop("method")
        except KeyError:
            raise ValueError("call object must have a method") from None
        if not isinstance(method, str):
            raise TypeError(
                f"method field of call object must be a str, is "
                f"{method.__class__.__name__}"
            )
        parameters = extensions.pop("parameters", {})
        if not isinstance(parameters, dict):
            raise TypeError(
                f"call parameters must be map, are "
                f"{parameters.__class__.__name__}"
            )
        oneway = extensions.pop("oneway", False)
        if not isinstance(oneway, bool):
            raise TypeError(
                f"call property oneway must be bool, is "
                f"{oneway.__class__.__name__}"
            )
        more = extensions.pop("more", False)
        if not isinstance(more, bool):
            raise TypeError(
                f"call property more must be bool, is "
                f"{more.__class__.__name__}"
            )
        upgrade = extensions.pop("upgrade", False)
        if not isinstance(upgrade, bool):
            raise TypeError(
                f"call property upgrade must be bool, is "
                f"{upgrade.__class__.__name__}"
            )
        if sum((oneway, more, upgrade)) > 1:
            raise ValueError("cannot combine oneway, more or upgrade")
        return cls(
            method,
            parameters,
            oneway,
            more,
            upgrade,
            extensions,
        )

    def tojson(self) -> JSONObject:
        """Export as a JSONObject suitable for json.dumps."""
        result: JSONObject = {"method": self.method}
        if self.parameters:
            result["parameters"] = self.parameters
        if self.oneway:
            result["oneway"] = True
        if self.more:
            result["more"] = True
        if self.upgrade:
            result["upgrade"] = True
        result.update(self.extensions)
        return result


@dataclasses.dataclass
class VarlinkMethodReply:
    """Represent a parsed and roughly validated varlink method reply."""

    parameters: JSONObject
    continues: bool = False
    error: str | None = None
    extensions: JSONObject = dataclasses.field(default_factory=dict)

    @property
    def error_interface(self) -> str:
        """Return the interface portion of the error string if any. May raise a
        ValueError.
        """
        if self.error is None:
            raise ValueError("not an error")
        interface, dot, _ = self.error.rpartition(".")
        if dot != ".":
            raise ValueError("unqualified error string")
        return interface

    @property
    def error_name(self) -> str:
        """Return the unqualified name portion of the error string if any. May
        raise a ValueError.
        """
        if self.error is None:
            raise ValueError("not an error")
        return self.error.rpartition(".")[2]

    def __post_init__(self) -> None:
        if self.error is not None:
            validate_interface(self.error_interface)
            validate_name(self.error_name)

    @classmethod
    def fromjson(cls, obj: JSONValue) -> "VarlinkMethodReply":
        """Parse a JSON value into a validated VarlinkMethodReply. May raise
        TypeError and ValueError.
        """
        if not isinstance(obj, dict):
            raise TypeError(
                f"call object must be a map, is {obj.__class__.__name__}"
            )
        extensions = obj.copy()
        parameters = extensions.pop("parameters", {})
        if not isinstance(parameters, dict):
            raise TypeError(
                f"reply parameters must be map, are "
                f"{parameters.__class__.__name__}"
            )
        continues = extensions.pop("continues", False)
        if not isinstance(continues, bool):
            raise TypeError(
                f"reply property continues must be bool, is "
                f"{continues.__class__.__name__}"
            )
        error = extensions.pop("error", None)
        if error is not None and not isinstance(error, str):
            raise TypeError(
                f"reply property error must be str, is "
                f"{error.__class__.__name__}"
            )
        return VarlinkMethodReply(parameters, continues, error, extensions)

    def tojson(self) -> JSONObject:
        """Export as a JSONObject suitable for json.dumps."""
        result: JSONObject = {}
        if self.continues:
            result["continues"] = True
        if self.error:
            result["error"] = self.error
        if self.parameters:
            result["parameters"] = self.parameters
        result.update(self.extensions)
        return result



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""A varlink implementation in pure Python with a few key design choices:

* asyncio: There is no synchronous support.
* file descriptor passing: Even though the varlink faq says that passing file
  descriptors is out of scope, systemd does this and it and this library
  supports such use.
* automatic introspection via type annotations: Rather than having to write a
  .varlink description file supporting introspection. This is being computed
  from Python type annotations.
"""

from .clientprotocol import *
from .conversion import *
from .error import *
from .interface import *
from .message import *
from .protocol import *
from .serverprotocol import *
from .types import *
from .util import *



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Helper for converting between Python objects and JSONValues."""

import collections.abc
import contextlib
import dataclasses
import enum
import types
import typing

from .types import (
    FileDescriptor,
    FileDescriptorArray,
    JSONValue,
    JSONObject,
    override,
)


class ConversionError(Exception):
    """A failure to convert a Python value from or to a JSONValue."""

    def __init__(self, message: str, location: list[str | int] | None = None):
        self.message = message
        self.location: list[str | int] = [] if location is None else location

    @classmethod
    def expected(
        cls,
        what: str,
        obj: object,
        location: list[str | int] | None = None,
    ) -> "ConversionError":
        """Construct a Conversion error indicating that something described by
        what was expected whereas a different type was found.
        """
        return cls(
            f"expected {what}, but got a {type(obj).__name__}", location
        )

    @classmethod
    @contextlib.contextmanager
    def context(cls, where: str | int) -> collections.abc.Iterator[None]:
        """If a ConversionError passes through this context manager, push the
        location where onto the location stack as it passes through.
        """
        try:
            yield
        except ConversionError as err:
            err.location.insert(0, where)
            raise


OOBTypeState = collections.abc.Mapping[type["VarlinkType"], typing.Any] | None


class VarlinkType:
    """A type abstraction that is exportable simultaneously to Python type
    annotations and varlink interface descriptions.
    """

    as_type: typing.Any
    """A Python type representing the varlink type suitable for a type
    annotation.
    """

    as_varlink: str
    """A varlink interface description representation of the varlink type."""

    typedefs: dict[str, str] = {}
    """Varlink type definitions used in the varlink type."""

    contains_fds: bool
    """Indicate whether fds requiring lifetime management are contained
    somewhere inside this type.
    """

    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONValue:
        """Convert a Python object conforming to the as_type type annotation to
        a json-convertible object suitable for consumption by varlink. A
        conversion may use the optional out-of-band state object using its own
        type as key and should otherwise forward the oobstate during recursion.
        """
        raise NotImplementedError

    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        """Convert a json-decoded Python object to a Python object conforming
        to the as_type type annotation. A conversion may use the optional
        out-of-band state object using its own type as key and should otherwise
        forward the oobstate during recursion.
        """
        raise NotImplementedError

    @classmethod
    def from_type_annotation(cls, tobj: typing.Any) -> "VarlinkType":
        """Convert a Python type annotation object into the VarlinkType
        abstraction. Note that this conversion is lossy and will convert
        unknown types to typing.Any/"object".
        """
        origin = typing.get_origin(tobj)
        args = typing.get_args(tobj)
        if origin is None:
            if isinstance(tobj, type):
                if issubclass(tobj, bool):
                    return SimpleVarlinkType("bool", bool)
                if issubclass(tobj, int):
                    return SimpleVarlinkType("int", int)
                if issubclass(tobj, float):
                    return SimpleVarlinkType("float", float, int)
                if issubclass(tobj, str):
                    return SimpleVarlinkType("string", str)
                if issubclass(tobj, enum.Enum):
                    return EnumVarlinkType(tobj)
                if issubclass(tobj, FileDescriptor):
                    return FileDescriptorVarlinkType()
            if typing.is_typeddict(tobj):
                # Do not iterate __*_keys__ as their order is unstable.
                return ObjectVarlinkType(
                    {
                        name: (
                            cls.from_type_annotation(elemtype).optional()
                            if name in tobj.__optional_keys__
                            else cls.from_type_annotation(elemtype)
                        )
                        for name, elemtype in tobj.__annotations__.items()
                    },
                )
            if dataclasses.is_dataclass(tobj):
                # is_dataclass also returns True for instances.
                assert isinstance(tobj, type)
                return DataclassVarlinkType(tobj)
        elif origin is typing.Literal:
            if all(isinstance(arg, str) for arg in args):
                return LiteralVarlinkType(args)
        elif origin is typing.Union or origin is types.UnionType:
            if any(arg is types.NoneType for arg in args):
                remaining = [alt for alt in args if alt is not types.NoneType]
                if remaining:
                    if len(remaining) == 1:
                        result = cls.from_type_annotation(remaining[0])
                    else:
                        result = cls.from_type_annotation(
                            typing.Union[tuple(remaining)]
                        )
                    if isinstance(
                        result, (ForeignVarlinkType, OptionalVarlinkType)
                    ):
                        return result
                    return OptionalVarlinkType(result)
        elif origin is list:
            if len(args) == 1:
                return ListVarlinkType(cls.from_type_annotation(args[0]))
        elif origin is dict:
            if len(args) == 2 and args[0] is str:
                return DictVarlinkType(cls.from_type_annotation(args[1]))
        elif origin is set:
            if len(args) == 1 and args[0] is str:
                return SetVarlinkType()
        return ForeignVarlinkType()

    def optional(self) -> "OptionalVarlinkType":
        """Wrap the type representation in an OptionalVarlinkType."""
        return OptionalVarlinkType(self)


def _merge_typedefs(
    typedefs: dict[str, str], *moretypedefs: dict[str, str]
) -> None:
    """Update the given typedefs in-place effectively computing a union of all
    given moretypedefs. Conflicting definitions are rejected with a
    RuntimeError.
    """
    for defmap in moretypedefs:
        for name, tdef in defmap.items():
            try:
                olddef = typedefs[name]
            except KeyError:
                typedefs[name] = tdef
            else:
                if olddef != tdef:
                    raise RuntimeError(
                        "conflicting type definitions for "
                        f"{name}: {olddef} vs {tdef}"
                    )


class SimpleVarlinkType(VarlinkType):
    """A varlink type representing a base type such as int or str."""

    contains_fds = False

    def __init__(self, varlinktype: str, pythontype: type, *convertible: type):
        self.as_type = pythontype
        self.as_varlink = varlinktype
        self._convertible = tuple(convertible)

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONValue:
        if isinstance(obj, self.as_type):
            return typing.cast(JSONValue, obj)
        if isinstance(obj, self._convertible):
            try:
                return typing.cast(JSONValue, self.as_type(obj))
            except Exception as exc:
                raise ConversionError(
                    f"expected {self.as_varlink}, but failed to convert from "
                    f"{type(obj).__name__}"
                ) from exc
        raise ConversionError.expected(self.as_varlink, obj)

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if isinstance(obj, self.as_type):
            return obj
        if isinstance(obj, self._convertible):
            try:
                return self.as_type(obj)
            except Exception as exc:
                raise ConversionError(
                    f"expected {self.as_varlink}, but failed to convert from "
                    f"{type(obj).__name__}"
                ) from exc
        raise ConversionError.expected(self.as_varlink, obj)

    def __repr__(self) -> str:
        typestr = ", ".join(
            tobj.__name__ if tobj in {bool, float, int, str} else repr(tobj)
            for tobj in (self.as_type,) + self._convertible
        )
        return f"{self.__class__.__name__}({self.as_varlink!r}, {typestr})"


class OptionalVarlinkType(VarlinkType):
    """A varlink type that allows an optional null/None value."""

    def __init__(self, vtype: VarlinkType):
        if isinstance(vtype, OptionalVarlinkType):
            raise RuntimeError("cannot nest OptionalVarlinkTypes")
        self._vtype = vtype
        self.as_type = vtype.as_type | None
        self.as_varlink = "?" + vtype.as_varlink
        self.typedefs = vtype.typedefs
        self.contains_fds = vtype.contains_fds

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONValue:
        if obj is None:
            return None
        return self._vtype.tojson(obj, oobstate)

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if obj is None:
            return None
        return self._vtype.fromjson(obj, oobstate)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._vtype!r})"

    @override
    def optional(self) -> typing.Self:
        return self


class ListVarlinkType(VarlinkType):
    """A varlink type representing a homogeneous array/list value."""

    def __init__(self, elttype: VarlinkType):
        self._elttype = elttype
        # mypy cannot runtime-constructed type hints.
        self.as_type = list[elttype.as_type]  # type: ignore[name-defined]
        self.as_varlink = "[]" + elttype.as_varlink
        self.typedefs = elttype.typedefs
        self.contains_fds = elttype.contains_fds

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> list[JSONValue]:
        if not isinstance(obj, list):
            raise ConversionError.expected("list", obj)
        result: list[JSONValue] = []
        for elt in obj:
            with ConversionError.context(len(result)):
                result.append(self._elttype.tojson(elt, oobstate))
        return result

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if not isinstance(obj, list):
            raise ConversionError.expected("list", obj)
        result: list[typing.Any] = []
        for elt in obj:
            with ConversionError.context(len(result)):
                result.append(self._elttype.fromjson(elt, oobstate))
        return result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._elttype!r})"


class DictVarlinkType(VarlinkType):
    """A varlink type representing a map/dict with string keys and homogeneous
    value types.
    """

    def __init__(self, elttype: VarlinkType):
        self._elttype = elttype
        # mypy cannot runtime-constructed type hints.
        self.as_type = dict[str, elttype.as_type]  # type: ignore[name-defined]
        self.as_varlink = "[string]" + elttype.as_varlink
        self.typedefs = elttype.typedefs
        self.contains_fds = elttype.contains_fds

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONObject:
        if not isinstance(obj, dict):
            raise ConversionError.expected("dict", obj)
        result = {}
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ConversionError.expected("str as dict key", key)
            with ConversionError.context(key):
                result[key] = self._elttype.tojson(value, oobstate)
        return result

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if not isinstance(obj, dict):
            raise ConversionError.expected("map", obj)
        result = {}
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ConversionError.expected("string as map key", key)
            with ConversionError.context(key):
                result[key] = self._elttype.fromjson(value, oobstate)
        return result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._elttype!r})"


class SetVarlinkType(VarlinkType):
    """A varlink type that represents a set of strings as a map on the varlink
    side.
    """

    as_type = set[str]
    as_varlink = "[string]()"
    contains_fds = False

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONObject:
        if not isinstance(obj, set):
            raise ConversionError.expected("set", obj)
        result: dict[str, JSONValue] = {}
        for elem in obj:
            if not isinstance(elem, str):
                raise ConversionError.expected("str as set element", elem)
            # Assign an empty struct as value.
            result[elem] = {}
        return result

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if not isinstance(obj, dict):
            raise ConversionError.expected("map", obj)
        result = set()
        for key, value in obj.items():
            if not isinstance(key, str):
                raise ConversionError.expected("string as map key", key)
            if value != {}:
                raise ConversionError.expected(
                    "empty struct as map value", value, [key]
                )
            result.add(key)
        return result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class ObjectVarlinkType(VarlinkType):
    """A varlink type representing a map/dict with string keys and
    inhomogeneous per-key types. On the Python side, this resembles a
    typing.TypedDict.
    """

    def __init__(
        self,
        typemap: dict[str, VarlinkType],
    ):
        self._typemap = typemap
        # mypy cannot runtime-constructed type hints.
        self.as_type = typing.TypedDict(  # type: ignore[misc]
            "ObjectVarlinkTypedDict",
            {name: tobj.as_type for name, tobj in typemap.items()},
        )
        self.as_varlink = "(%s)" % ", ".join(
            f"{name}: {tobj.as_varlink}" for name, tobj in typemap.items()
        )
        self.typedefs = {}
        _merge_typedefs(
            self.typedefs, *(tobj.typedefs for tobj in typemap.values())
        )
        self.contains_fds = any(tobj.contains_fds for tobj in typemap.values())

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONObject:
        if not isinstance(obj, dict):
            raise ConversionError.expected("dict", obj)
        result = {}
        for key, vtype in self._typemap.items():
            try:
                value = obj[key]
            except KeyError as err:
                if isinstance(vtype, OptionalVarlinkType):
                    continue
                raise ConversionError(
                    f"missing required key {key} in given dict", [key]
                ) from err
            with ConversionError.context(key):
                result[key] = vtype.tojson(value, oobstate)
        for key in obj:
            if key not in self._typemap:
                raise ConversionError(f"no type for key {key}", [key])
        return result

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if not isinstance(obj, dict):
            raise ConversionError.expected("map", obj)
        result = {}
        for key, vtype in self._typemap.items():
            try:
                value = obj[key]
            except KeyError as err:
                if isinstance(vtype, OptionalVarlinkType):
                    continue
                raise ConversionError(
                    f"missing required key {key} in given dict"
                ) from err
            with ConversionError.context(key):
                result[key] = vtype.fromjson(value, oobstate)
        for key in obj:
            if key not in self._typemap:
                raise ConversionError(f"no type for key {key}", [key])
        return result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._typemap!r})"


class DataclassVarlinkType(VarlinkType):
    """A varlink type representing an object as a dataclasses.dataclass.
    Only fields with init=True will be considered."""

    def __init__(self, dataclasstype: type):
        assert dataclasses.is_dataclass(dataclasstype)
        self.as_type = dataclasstype
        self.as_varlink = dataclasstype.__name__
        self._typemap = {
            field.name: VarlinkType.from_type_annotation(field.type)
            for field in dataclasses.fields(dataclasstype)
            if field.init
        }
        self.typedefs = {}
        _merge_typedefs(
            self.typedefs,
            *(vtype.typedefs for vtype in self._typemap.values()),
            {
                dataclasstype.__name__: "(%s)"
                % ", ".join(
                    f"{name}: {vtype.as_varlink}"
                    for name, vtype in self._typemap.items()
                ),
            },
        )
        self.contains_fds = any(
            tobj.contains_fds for tobj in self._typemap.values()
        )

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONObject:
        if not dataclasses.is_dataclass(obj):
            raise ConversionError.expected("a dataclass", obj)
        result = {}
        for name, vtype in self._typemap.items():
            with ConversionError.context(name):
                result[name] = vtype.tojson(getattr(obj, name), oobstate)
        return result

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if not isinstance(obj, dict):
            raise ConversionError.expected("map", obj)
        excess_keys = set(obj.keys())
        excess_keys.difference_update(self._typemap.keys())
        if excess_keys:
            raise ConversionError(
                "unexpected dataclass fields " + ", ".join(excess_keys)
            )
        fields = {}
        for name, vtype in self._typemap.items():
            try:
                value = obj[name]
            except KeyError as err:
                if not isinstance(vtype, OptionalVarlinkType):
                    raise ConversionError(
                        f"missing object key {name}"
                    ) from err
                value = None
            with ConversionError.context(name):
                fields[name] = vtype.fromjson(value, oobstate)
        return self.as_type(**fields)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.as_type})"


class EnumVarlinkType(VarlinkType):
    """A varlink type representing an enum as an enum.Enum."""

    contains_fds = False

    def __init__(self, enumtype: type[enum.Enum]) -> None:
        if not issubclass(enumtype, enum.Enum):
            raise TypeError("a subclass of Enum is required")
        self.as_type = enumtype
        self.as_varlink = enumtype.__name__
        self.typedefs = {
            enumtype.__name__: "(%s)" % ", ".join(enumtype.__members__)
        }

    @override
    def tojson(self, obj: typing.Any, oobstate: OOBTypeState = None) -> str:
        if not isinstance(obj, self.as_type):
            raise ConversionError.expected(f"enum {self.as_type!r}", obj)
        assert isinstance(obj, enum.Enum)
        return obj.name

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if not isinstance(obj, str):
            raise ConversionError.expected("string as enum value", obj)
        try:
            return self.as_type.__members__[obj]
        except KeyError as err:
            raise ConversionError(
                f"enum {self.as_type!r} value {obj!r} not known"
            ) from err

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.as_type!r})"


class LiteralVarlinkType(VarlinkType):
    """A varlink type representation for an enum represented as
    typing.Literal.
    """

    contains_fds = False

    def __init__(self, values: tuple[str, ...]):
        self._values = values
        # mypy cannot handle dynamic literals
        self.as_type = typing.Literal[values]
        self.as_varlink = "(%s)" % ", ".join(values)

    @override
    def tojson(self, obj: typing.Any, oobstate: OOBTypeState = None) -> str:
        if not (isinstance(obj, str) and obj in self._values):
            raise ConversionError(f"invalid literal value {obj!r}")
        return obj

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        if not (isinstance(obj, str) and obj in self._values):
            raise ConversionError(f"invalid literal value {obj!r}")
        return obj

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._values!r})"


class FileDescriptorVarlinkType(VarlinkType):
    """Represent a file descriptor in a varlink message. On the specification
    side, this use is explicitly ruled out. systemd does it anyway and so
    does this class.
    """

    as_type = FileDescriptor
    as_varlink = "int"
    contains_fds = True

    @classmethod
    def _get_oob(cls, oobstate: OOBTypeState) -> typing.Any:
        if oobstate is None:
            raise ConversionError(
                "cannot convert a file descriptor without oobstate"
            )
        try:
            return oobstate[cls]
        except KeyError:
            raise ConversionError(
                "cannot convert a file descriptor without associated oobstate"
            ) from None

    @override
    def tojson(self, obj: typing.Any, oobstate: OOBTypeState = None) -> int:
        """Represent a file descriptor. It may be conveyed as int | HasFileno.
        The actual file descriptor is appended to the out-of-band state array
        and the returned json value is the index into said array. A
        FileDescriptorArray should be conveyed as out-of-band state.
        """
        if not (isinstance(obj, int) or hasattr(obj, "fileno")):
            raise ConversionError.expected("int or fileno()-like", obj)
        fdarray = self._get_oob(oobstate)
        if not isinstance(fdarray, FileDescriptorArray):
            raise ConversionError(
                f"out-of-band state for {self.__class__.__name__} should be "
                f"FileDescriptorArray, is {type(fdarray)}"
            )
        try:
            index = fdarray.add(obj)
        except ValueError as err:
            raise ConversionError("invalid file descriptor") from err
        return index

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        """Unrepresent a file descriptor. The int value is understood as an
        index into the out-of-band state object of type FileDescriptorArray. A
        file descriptor is looked up at the index and the position is released
        there.
        """
        if not isinstance(obj, int):
            raise ConversionError.expected("int", obj)
        if oobstate is None:
            raise ConversionError(
                "cannot unrepresent a file descriptor without oobstate"
            )
        fds = self._get_oob(oobstate)
        if not isinstance(fds, FileDescriptorArray):
            raise ConversionError(
                f"out-of-band state for {self.__class__.__name__} should be "
                f"FileDescriptorArray, is {type(fds)}"
            )
        try:
            return fds[obj]
        except IndexError as err:
            raise ConversionError(
                f"attempt to convert invalid file descriptor index {obj}"
            ) from err


class ForeignVarlinkType(VarlinkType):
    """A varlink type skipping representing a foreign object or typing.Any
    and skipping value conversion steps.
    """

    as_type = typing.Any
    contains_fds = False

    def __init__(self, varlink: str = "object"):
        self.as_varlink = varlink

    @override
    def tojson(
        self, obj: typing.Any, oobstate: OOBTypeState = None
    ) -> JSONValue:
        # We have no guarantuee that the object actually is a JSONValue and
        # hope that the user is doing things correctly.
        return typing.cast(JSONValue, obj)

    @override
    def fromjson(
        self, obj: JSONValue, oobstate: OOBTypeState = None
    ) -> typing.Any:
        return obj



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Python classes for errors of the org.varlink.service interface."""

from .error import TypedVarlinkErrorReply


class InterfaceNotFound(
    TypedVarlinkErrorReply, interface="org.varlink.service"
):
    """A method was called for an interface that is not provided or a
    method was called with an interface parameter and said interface is not
    provided.
    """

    class Parameters:
        interface: str


class MethodNotFound(TypedVarlinkErrorReply, interface="org.varlink.service"):
    """A method was called that is not provided on the named interface."""

    class Parameters:
        method: str


class MethodNotImplemented(
    TypedVarlinkErrorReply, interface="org.varlink.service"
):
    """The interface defines the requested method, but the service does not
    implement it.
    """

    class Parameters:
        method: str


class InvalidParameter(
    TypedVarlinkErrorReply, interface="org.varlink.service"
):
    """A call parameter could not be validated to the expected type."""

    class Parameters:
        parameter: str


class PermissionDenied(
    TypedVarlinkErrorReply, interface="org.varlink.service"
):
    """The service refuses to service the request as it denies access."""

    class Parameters:
        pass


class ExpectedMore(TypedVarlinkErrorReply, interface="org.varlink.service"):
    """A method was expecting to be called with "more": true, but was called
    without.
    """

    class Parameters:
        pass



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import dataclasses
import enum
import typing
import unittest

import hypothesis
import hypothesis.strategies as st

from asyncvarlink import (
    ConversionError,
    DataclassVarlinkType,
    DictVarlinkType,
    EnumVarlinkType,
    FileDescriptor,
    FileDescriptorArray,
    FileDescriptorVarlinkType,
    ForeignVarlinkType,
    JSONValue,
    ListVarlinkType,
    LiteralVarlinkType,
    ObjectVarlinkType,
    OptionalVarlinkType,
    SetVarlinkType,
    SimpleVarlinkType,
    VarlinkType,
)


@dataclasses.dataclass
class Quantity:
    value: float
    unit: str


class TriState(enum.Enum):
    rock = "rock"
    paper = "paper"
    scissors = "scissors"


class Digits(enum.Enum):
    zero = "zero"
    one = "one"
    two = "two"
    three = "three"
    four = "four"
    five = "five"
    six = "six"
    seven = "seven"
    eight = "eight"
    nine = "nine"


class Dummy:
    pass


@st.deferred
def type_annotations() -> st.SearchStrategy[type]:
    return st.one_of(
        st.just(bool),
        st.just(int),
        st.just(FileDescriptor),
        st.just(float),
        st.just(Quantity),
        st.just(str),
        st.just(TriState),
        st.just(typing.Literal["on", "off"]),
        st.just(Digits),
        st.just(set[str]),
        st.builds(
            lambda fields: typing.TypedDict(
                "AnonTypedDict",
                {
                    key: ta if required else typing.NotRequired[ta]
                    for key, (ta, required) in fields.items()
                },
            ),
            st.dictionaries(
                st.text(),
                st.tuples(type_annotations, st.booleans()),
            ),
        ),
        st.builds(lambda ta: ta | None, type_annotations),
        st.builds(lambda ta: list[ta], type_annotations),
        st.builds(lambda ta: dict[str, ta], type_annotations),
        st.builds(lambda: Dummy),
    )


class MockedFd:
    """Looks like file descriptor, but isn't and swallows close."""

    def __init__(self, fd: int, check_del: bool = False):
        self.fd = fd
        self.check_del = check_del

    def fileno(self) -> int:
        return self.fd

    def close(self) -> None:
        assert self.fd >= 0
        self.fd = -1

    def __del__(self) -> None:
        assert self.fd < 0 or not self.check_del


# Precreate some fds and do not check their __del__.
mocked_fds = [MockedFd(n) for n in range(10)]


def representable(vt: VarlinkType) -> st.SearchStrategy[typing.Any]:
    if isinstance(vt, (SimpleVarlinkType, EnumVarlinkType)):
        if vt.as_type == float:
            return st.floats(allow_nan=False)
        return st.from_type(vt.as_type)
    if isinstance(vt, LiteralVarlinkType):
        return st.sampled_from(sorted(vt._values))
    if isinstance(vt, FileDescriptorVarlinkType):
        return st.sampled_from(mocked_fds)
    if isinstance(vt, OptionalVarlinkType):
        return st.one_of(st.none(), representable(vt._vtype))
    if isinstance(vt, ListVarlinkType):
        return st.lists(representable(vt._elttype))
    if isinstance(vt, DictVarlinkType):
        return st.dictionaries(st.text(), representable(vt._elttype))
    if isinstance(vt, SetVarlinkType):
        return st.sets(st.text())
    if isinstance(vt, ObjectVarlinkType):
        return st.builds(
            dict,
            **{
                key: representable(value)
                for key, value in vt._typemap.items()
                if not isinstance(value, OptionalVarlinkType)
            },
        )
    if isinstance(vt, DataclassVarlinkType):
        assert vt.as_type is Quantity
        return st.builds(
            Quantity,
            value=st.floats(allow_nan=False),
            unit=st.from_type(str),
        )
    assert isinstance(vt, ForeignVarlinkType)
    return st.just(object())


json_values = st.recursive(
    st.one_of(
        st.none(),
        st.booleans(),
        st.integers(),
        st.floats(allow_nan=False),
        st.text(),
    ),
    lambda children: st.lists(children) | st.dictionaries(st.text(), children),
)


class Coordinate(typing.TypedDict):
    x: float
    y: float


def close_all_fds(obj: typing.Any) -> None:
    if isinstance(obj, FileDescriptor):
        obj.close()
    elif isinstance(obj, list):
        for elem in obj:
            close_all_fds(elem)
    elif isinstance(obj, dict):
        for elem in obj.values():
            close_all_fds(elem)


class ConversionTests(unittest.TestCase):
    @hypothesis.given(type_annotations, st.data())
    def test_round_trip(self, ta: type, data) -> None:
        vt = VarlinkType.from_type_annotation(ta)
        obj = data.draw(representable(vt))
        sentinel = object()
        fdarray = FileDescriptorArray(initial_referee=sentinel)
        oobto: dict[type, typing.Any] = {FileDescriptorVarlinkType: fdarray}
        val = vt.tojson(obj, oobto)
        # We don't have to dispose obj as its MockedFds don't check __del__,
        # but we want the array to be checked.
        for fd in fdarray:
            self.assertIsInstance(fd, FileDescriptor)
            self.assertIsInstance(fd.fd, MockedFd)
            assert isinstance(fd.fd, MockedFd)  # help mypy
            fd.fd = MockedFd(fd.fd.fd, check_del=True)
        oobfrom: dict[type, typing.Any] = {FileDescriptorVarlinkType: fdarray}
        obj_again = vt.fromjson(val, oobfrom)
        self.assertEqual(obj, obj_again)
        close_all_fds(obj_again)
        self.assertFalse(bool(fdarray))
        fdarray.release(sentinel)

    @hypothesis.given(type_annotations, json_values)
    def test_exception(self, ta: type, val: JSONValue) -> None:
        vt = VarlinkType.from_type_annotation(ta)
        try:
            vt.fromjson(val)
        except ConversionError:
            pass

    def test_invalid(self) -> None:
        testcases: list[tuple[type, list[JSONValue], list[typing.Any]]] = [
            (bool, ["fuzzy"], ["fuzzy"]),
            (int, [2.5], [2.5]),
            (float, ["fuzzy", 1 << 9999], ["fuzzy"]),
            (str, [42], [42]),
            (list[str], [2, [2]], [2, [2]]),
            (set[str], [2, {"x": 1}], [1, {1}]),
            (dict[str, str], [2, {"x": 1}], [1, [("x", "y")], {1: 2}]),
            (
                Coordinate,
                [
                    2,
                    {"x": 1.0},
                    {"x": 1, "y": "bad"},
                    {"x": 1.0, "y": 1.0, "z": 1.0},
                ],
                [
                    2,
                    {"x": 1.0},
                    {"x": 1, "y": "bad"},
                    {"x": 1.0, "y": 1.0, "z": 1.0},
                ],
            ),
            (TriState, [2, "well"], [2, "well"]),
            (typing.Literal["on", "off"], [2, "maybe"], [2, "maybe"]),
            (FileDescriptor, ["stdin", -1], ["stdin"]),
        ]
        for ta, jvals, pvals in testcases:
            oobstate: dict[type[VarlinkType], typing.Any]
            vt = VarlinkType.from_type_annotation(ta)
            for jval in jvals:
                oobstate = {
                    FileDescriptorVarlinkType: FileDescriptorArray(self),
                }
                with self.subTest(ta=ta, jval=jval):
                    with self.assertRaises(ConversionError):
                        vt.fromjson(jval, oobstate=oobstate)
            for pval in pvals:
                with self.subTest(ta=ta, pval=pval):
                    oobstate = {FileDescriptorVarlinkType: []}
                    with self.assertRaises(ConversionError):
                        vt.tojson(pval, oobstate=oobstate)



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

import os
import unittest
import unittest.mock

from asyncvarlink import get_listen_fd


class GetListenFdTests(unittest.TestCase):
    def test_missing(self) -> None:
        pid = os.getpid()
        with unittest.mock.patch.dict(
            "os.environ", {"LISTEN_FDS": "1"}, clear=True
        ):
            self.assertIsNone(get_listen_fd("spam"), "missing LISTEN_PID")
        with unittest.mock.patch.dict(
            "os.environ",
            {"LISTEN_FDS": "1", "LISTEN_PID": str(pid + 1)},
            clear=True,
        ):
            self.assertIsNone(get_listen_fd("spam"), "wrong LISTEN_PID")
        with unittest.mock.patch.dict(
            "os.environ",
            {"LISTEN_FDS": "2", "LISTEN_PID": str(pid)},
            clear=True,
        ):
            self.assertIsNone(get_listen_fd("spam"), "missing LISTEN_FDNAMES")
        with unittest.mock.patch.dict(
            "os.environ",
            {
                "LISTEN_FDS": "1",
                "LISTEN_PID": str(pid),
                "LISTEN_FDNAMES": "beacon",
            },
            clear=True,
        ):
            self.assertIsNone(get_listen_fd("spam"), "mismatch LISTEN_FDNAMES")
        with unittest.mock.patch.dict(
            "os.environ",
            {
                "LISTEN_FDS": "2",
                "LISTEN_PID": str(pid),
                "LISTEN_FDNAMES": "beacon",
            },
            clear=True,
        ):
            self.assertIsNone(get_listen_fd("spam"), "bad LISTEN_FDNAMES")
        with unittest.mock.patch.dict(
            "os.environ",
            {
                "LISTEN_FDS": "2",
                "LISTEN_PID": str(pid),
                "LISTEN_FDNAMES": "beacon:egg",
            },
            clear=True,
        ):
            self.assertIsNone(
                get_listen_fd("spam"), "name not in LISTEN_FDNAMES"
            )

    def test_found(self) -> None:
        pid = str(os.getpid())
        with unittest.mock.patch.dict(
            "os.environ", {"LISTEN_FDS": "1", "LISTEN_PID": pid}, clear=True
        ):
            # If there is only one fd and no LISTEN_FDNAMES, expect success for
            # backwards compatibility.
            fd = get_listen_fd("spam")
            self.assertEqual(fd, 3)
            assert fd is not None  # help mypy
            fd.take()
        with unittest.mock.patch.dict(
            "os.environ",
            {"LISTEN_FDS": "1", "LISTEN_PID": pid, "LISTEN_FDNAMES": "spam"},
            clear=True,
        ):
            fd = get_listen_fd("spam")
            self.assertEqual(fd, 3)
            assert fd is not None  # help mypy
            fd.take()
        with unittest.mock.patch.dict(
            "os.environ",
            {
                "LISTEN_FDS": "2",
                "LISTEN_PID": pid,
                "LISTEN_FDNAMES": "baecon:spam",
            },
            clear=True,
        ):
            fd = get_listen_fd("spam")
            self.assertEqual(fd, 4)
            assert fd is not None  # help mypy
            fd.take()



# Copyright 2024 Helmut Grohne <helmut@subdivi.de>
# SPDX-License-Identifier: LGPL-2.0-or-later

"""Utility functions that don't fit well elsewhere."""

import asyncio
import collections.abc
import contextlib
import os
import socket
import stat
import typing
import weakref

from .protocol import _BLOCKING_ERRNOS, VarlinkBaseProtocol, VarlinkTransport
from .types import FileDescriptor, override


_T = typing.TypeVar("_T")


@typing.overload
def completing_future() -> (
    contextlib.AbstractContextManager[asyncio.Future[None]]
): ...


@typing.overload
def completing_future(
    value: _T,
) -> contextlib.AbstractContextManager[asyncio.Future[_T]]: ...


@contextlib.contextmanager
def completing_future(
    value: typing.Any = None,
) -> collections.abc.Iterator[asyncio.Future[typing.Any]]:
    """A context manager that returns a new asyncio.Future which will be done
    with the passed value on context exit unless the context exits with an
    exception in which case the future also raises the exception. Even though
    this is a synchronous context manager, it must be used in an asynchronous
    context.
    """
    future = asyncio.get_running_loop().create_future()
    done = False
    try:
        yield future
    except BaseException as exc:
        future.set_exception(exc)
        done = True
        raise
    finally:
        if not done:
            future.set_result(value)
        # The consumer of the exception typically is not interested in the
        # actual result, but Python may log an exception unless we retrieve it.
        future.exception()


_P = typing.TypeVar("_P", bound=VarlinkBaseProtocol)


async def connect_unix_varlink(
    protocol_factory: collections.abc.Callable[[], _P],
    path: os.PathLike[str] | str,
    *,
    loop: asyncio.AbstractEventLoop | None = None,
    inheritable: bool = False,
) -> tuple[VarlinkTransport, _P]:
    """Connect to the unix domain socket at given path and return a varlink
    connection.
    """
    if loop is None:
        loop = asyncio.get_running_loop()
    socktype = socket.SOCK_STREAM | socket.SOCK_NONBLOCK
    if not inheritable:
        socktype |= socket.SOCK_CLOEXEC
    sock = socket.socket(socket.AF_UNIX, socktype)
    try:
        await loop.sock_connect(sock, os.fspath(path))
    except:
        sock.close()
        raise
    protocol = protocol_factory()
    transport = VarlinkTransport(loop, sock, sock, protocol)
    await asyncio.sleep(0)  # wait for all call_soon
    return transport, protocol


class VarlinkUnixServer(asyncio.AbstractServer):
    """An asyncio server class that will construct VarlinkTransports for
    accepted connections.
    """

    def __init__(
        self,
        sock: socket.socket,
        protocol_factory: collections.abc.Callable[[], VarlinkBaseProtocol],
        *,
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        self._loop = asyncio.get_running_loop() if loop is None else loop
        self._sock: socket.socket | None = sock
        self._protocol_factory = protocol_factory
        self._serving: asyncio.Future[None] | None = None
        self._transports: weakref.WeakSet[VarlinkTransport] = weakref.WeakSet()

    @override
    def close(self) -> None:
        if self._serving is not None:
            assert self._sock is not None
            self._loop.remove_reader(self._sock)
            self._sock.close()
            self._sock = None
            self._serving.set_result(None)
            self._serving = None

    @override
    def get_loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @override
    def is_serving(self) -> bool:
        return self._serving is not None

    @override
    async def start_serving(self) -> None:
        if self._serving is None:
            if self._sock is None:
                raise ValueError("cannot start serving on closed server")
            self._serving = self._loop.create_future()
            self._sock.listen(1)
            self._loop.add_reader(self._sock, self._handle_accept)

    def _handle_accept(self) -> None:
        if self._sock is None:
            return  # closed before handler could run
        try:
            conn, addr = self._sock.accept()
        except OSError as err:
            if err.errno in _BLOCKING_ERRNOS:
                return
            self.close()
            raise
        conn.setblocking(False)
        self._transports.add(
            VarlinkTransport(
                self._loop,
                conn,
                conn,
                self._protocol_factory(),
                {"peername": addr},
            )
        )

    @override
    async def serve_forever(self) -> None:
        await self.start_serving()
        await self.wait_closed()

    @override
    async def wait_closed(self) -> None:
        if self._serving is not None:
            await self._serving
        for transport in self._transports.copy():
            try:
                await transport.closed_future
            except asyncio.CancelledError:
                pass

    @override
    def close_clients(self) -> None:
        for transport in self._transports.copy():
            transport.close()

    @override
    def abort_clients(self) -> None:
        for transport in self._transports.copy():
            transport.abort()


async def create_unix_server(
    protocol_factory: collections.abc.Callable[[], VarlinkBaseProtocol],
    path: os.PathLike[str] | str | None = None,
    *,
    loop: asyncio.AbstractEventLoop | None = None,
    sock: socket.socket | None = None,
    start_serving: bool = True,
    inheritable: bool | None = None,
) -> VarlinkUnixServer:
    """In a similar spirit to asyncio.SelectorEventLoop.create_unix_server
    create a UNIX domain socket server except that the transport class being
    used will be VarlinkTransport for being capable of sending and receiving
    file descriptors.

    Either the socket (sock) or path must be provided.
    """
    if sock is None:
        if path is None:
            raise ValueError("neither path nor sock specified")
        pathstr = os.fspath(path)
        socktype = socket.SOCK_STREAM | socket.SOCK_NONBLOCK
        if not inheritable:
            socktype |= socket.SOCK_NONBLOCK
        sock = socket.socket(socket.AF_UNIX, socktype)
        if not pathstr.startswith("\0"):
            try:
                st = os.stat(pathstr)
            except FileNotFoundError:
                pass
            else:
                if stat.S_ISSOCK(st.st_mode):
                    os.unlink(pathstr)
        try:
            sock.bind(pathstr)
        except:
            sock.close()
            raise
    elif inheritable is not None:
        raise ValueError("cannot specify both sock and inheritable")
    else:
        sock.setblocking(False)
    server = VarlinkUnixServer(sock, protocol_factory, loop=loop)
    if start_serving:
        await server.start_serving()
    return server


def get_listen_fd(name: str) -> FileDescriptor | None:
    """Obtain a file descriptor using the systemd socket activation
    protocol.
    """
    try:
        pid = int(os.environ["LISTEN_PID"])
        fds = int(os.environ["LISTEN_FDS"])
    except (KeyError, ValueError):
        return None
    if fds < 1 or pid != os.getpid():
        return None
    if fds == 1:
        if os.environ.get("LISTEN_FDNAMES", name) != name:
            return None
        return FileDescriptor(3, should_close=True)
    try:
        names = os.environ["LISTEN_FDNAMES"].split(":")
    except KeyError:
        return None
    if len(names) != fds:
        return None
    try:
        return FileDescriptor(3 + names.index(name), should_close=True)
    except ValueError:
        return None
